"""

gpu_test.py

Simple script to verify CUDA availability, used to verify a CUDA environment
for TF or PyTorch

"""

# Minimize TF printouts
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'

try:
    import logging
    logging.getLogger('tensorflow').setLevel(logging.ERROR)
except Exception:
    pass


#%% Torch/TF test functions

def torch_test():
    """
    Print diagnostic information about Torch/CUDA status, including Torch/CUDA versions
    and all available CUDA device names.

    Returns:
        int: The number of CUDA devices reported by PyTorch.
    """

    try:
        import torch
    except Exception as e: #noqa
        print('PyTorch unavailable, not running PyTorch tests.  PyTorch import error was:\n{}'.format(
            str(e)))
        return

    print('Torch version: {}'.format(str(torch.__version__)))
    print('CUDA available (according to PyTorch): {}'.format(torch.cuda.is_available()))
    if torch.cuda.is_available():
        print('CUDA version (according to PyTorch): {}'.format(torch.version.cuda))
        print('CuDNN version (according to PyTorch): {}'.format(torch.backends.cudnn.version()))

    device_ids = list(range(torch.cuda.device_count()))

    if len(device_ids) > 0:
        cuda_str = 'Found {} CUDA devices:'.format(len(device_ids))
        print(cuda_str)

        for device_id in device_ids:
            device_name = 'unknown'
            try:
                device_name = torch.cuda.get_device_name(device=device_id)
            except Exception as e: #noqa
                pass
            print('{}: {}'.format(device_id,device_name))
    else:
        print('No GPUs reported by PyTorch')

    try:
        if torch.backends.mps.is_built and torch.backends.mps.is_available():
            print('PyTorch reports that Metal Performance Shaders are available')
    except Exception:
        pass
    return len(device_ids)


def tf_test():
    """
    Print diagnostic information about TF/CUDA status.

    Returns:
        int: The number of CUDA devices reported by PyTorch.
    """

    try:
        import tensorflow as tf
    except Exception as e: #noqa
        print('TensorFlow unavailable, not running TF tests.  TF import error was:\n{}'.format(
            str(e)))
        return

    from tensorflow.python.platform import build_info as build
    print(f"TF version: {tf.__version__}")

    if 'cuda_version' not in build.build_info:
        print('TF does not appear to be built with CUDA')
    else:
        print(f"CUDA build version reported by TensorFlow: {build.build_info['cuda_version']}")
    if 'cudnn_version' not in build.build_info:
        print('TF does not appear to be built with CuDNN')
    else:
        print(f"CuDNN build version reported by TensorFlow: {build.build_info['cudnn_version']}")

    try:
        from tensorflow.python.compiler.tensorrt import trt_convert as trt
        print("Linked TensorRT version: {}".format(trt.trt_utils._pywrap_py_utils.get_linked_tensorrt_version()))
    except Exception:
        print('Could not probe TensorRT version')

    gpus = tf.config.list_physical_devices('GPU')
    if gpus is None:
        gpus = []

    if len(gpus) > 0:
        print('TensorFlow found the following GPUs:')
        for gpu in gpus:
            print(gpu.name)

    else:
        print('No GPUs reported by TensorFlow')

    return len(gpus)


#%% Command-line driver

if __name__ == '__main__':

    print('*** Running Torch tests ***\n')
    torch_test()

    print('\n*** Running TF tests ***\n')
    tf_test()



"""

validate_batch_results.py

Given a .json file containing MD results, validate that it's compliant with the format spec:

https://lila.science/megadetector-output-format

"""

#%% Constants and imports

import os
import sys
import json
import argparse

from tqdm import tqdm

from megadetector.detection.video_utils import is_video_file
from megadetector.utils.ct_utils import args_to_object, is_list_sorted # noqa

typical_info_fields = ['detector',
                       'detection_completion_time',
                       'classifier',
                       'classification_completion_time',
                       'detection_metadata',
                       'classifier_metadata']

required_keys = ['info',
                 'images',
                 'detection_categories']

typical_keys = ['classification_categories',
                'classification_category_descriptions']


#%% Classes

class ValidateBatchResultsOptions:
    """
    Options controlling the behavior of validate_bach_results()
    """

    def __init__(self):

        #: Should we verify that images exist?  If this is True, and the .json
        #: file contains relative paths, relative_path_base needs to be specified.
        self.check_image_existence = False

        #: If check_image_existence is True, where do the images live?
        #:
        #: If None, assumes absolute paths.
        self.relative_path_base = None

        #: Should we return the loaded data, or just the validation results?
        self.return_data = False

        #: Enable additional debug output
        self.verbose = False

        #: Should we raise errors immediately (vs. just catching and reporting)?
        self.raise_errors = False

# ...class ValidateBatchResultsOptions


#%% Main function

def validate_batch_results(json_filename,options=None):
    """
    Verify that [json_filename] is a valid MD output file.  Currently errors on invalid files.

    Args:
        json_filename (str): the filename to validate
        options (ValidateBatchResultsOptions, optional): all the parameters used to control this
            process, see ValidateBatchResultsOptions for details

    Returns:
        dict: a dict with a field called "validation_results", which is itself a dict.  The reason
        it's a dict inside a dict is that if return_data is True, the outer dict also contains all
        the loaded data.  The "validation_results" dict contains fields called "errors", "warnings",
        and "filename".  "errors" and "warnings" are lists of strings, although "errors" will never
        be longer than N=1, since validation fails at the first error.

    """

    if options is None:
        options = ValidateBatchResultsOptions()

    if options.verbose:
        print('Loading results from {}'.format(json_filename))

    with open(json_filename,'r') as f:
        d = json.load(f)

    validation_results = {}
    validation_results['filename'] = json_filename
    validation_results['warnings'] = []
    validation_results['errors'] = []

    if not isinstance(d,dict):

        validation_results['errors'].append('Input data is not a dict')
        to_return = {}
        to_return['validation_results'] = validation_results
        return to_return

    try:

        ## Info validation

        if 'info' not in d:
            raise ValueError('Input does not contain info field')

        info = d['info']

        if not isinstance(info,dict):
            raise ValueError('Input contains invalid info field')

        if 'format_version' not in info :
            raise ValueError('Input does not specify format version')

        format_version = float(info['format_version'])
        if format_version < 1.3:
            raise ValueError('This validator can only be used with format version 1.3 or later')


        ## Category validation

        if 'detection_categories' not in d:
            raise ValueError('Input does not contain detection_categories field')

        for k in d['detection_categories'].keys():
            # Category ID should be string-formatted ints
            if not isinstance(k,str):
                raise ValueError('Invalid detection category ID: {}'.format(k))
            _ = int(k)
            if not isinstance(d['detection_categories'][k],str):
                raise ValueError('Invalid detection category name: {}'.format(
                    d['detection_categories'][k]))

        if 'classification_categories' in d:
            for k in d['classification_categories'].keys():
                # Categories should be string-formatted ints
                if not isinstance(k,str):
                    raise ValueError('Invalid classification category ID: {}'.format(k))
                _ = int(k)
                if not isinstance(d['classification_categories'][k],str):
                    raise ValueError('Invalid classification category name: {}'.format(
                        d['classification_categories'][k]))


        ## Image validation

        if 'images' not in d:
            raise ValueError('images field not present')
        if not isinstance(d['images'],list):
            raise ValueError('Invalid images field')

        if options.verbose:
            print('Validating images')

        # im = d['images'][0]
        for i_im,im in tqdm(enumerate(d['images']),total=len(d['images']),disable=(not options.verbose)):

            if not isinstance(im,dict):
                raise ValueError('Invalid image at index {}'.format(i_im))
            if 'file' not in im:
                raise ValueError('Image without filename at index {}'.format(i_im))

            file = im['file']

            if 'detections' in im and im['detections'] is not None:

                for det in im['detections']:

                    assert 'category' in det, 'Image {} has a detection with no category'.format(file)
                    assert 'conf' in det, 'Image {} has a detection with no confidence'.format(file)
                    assert isinstance(det['conf'],float), \
                        'Image {} has an illegal confidence value'.format(file)
                    assert 'bbox' in det, 'Image {} has a detection with no box'.format(file)
                    assert det['category'] in d['detection_categories'], \
                        'Image {} has a detection with an unmapped category {}'.format(
                            file,det['category'])

                    if 'classifications' in det and det['classifications'] is not None:
                        for c in det['classifications']:
                            assert isinstance(c[0],str), \
                                'Image {} has an illegal classification category: {}'.format(file,c[0])
                            try:
                                _ = int(c[0])
                            except Exception:
                                raise ValueError('Image {} has an illegal classification category: {}'.format(
                                    file,c[0]))
                            assert isinstance(c[1],float) or isinstance(c[1], int)

                # ...for each detection

            # ...if this image has a detections field

            if options.check_image_existence:

                if options.relative_path_base is None:
                    file_abs = file
                else:
                    file_abs = os.path.join(options.relative_path_base,file)
                if not os.path.isfile(file_abs):
                    raise ValueError('Cannot find file {}'.format(file_abs))

            if 'failure' in im:
                if im['failure'] is not None:
                    if not isinstance(im['failure'],str):
                        raise ValueError('Image {} has an illegal [failure] value: {}'.format(
                            im['file'],str(im['failure'])))
                    if 'detections' not in im:
                        s = 'Image {} has a failure value, should also have a null detections array'.format(
                            im['file'])
                        validation_results['warnings'].append(s)
                    elif im['detections'] is not None:
                        raise ValueError('Image {} has a failure value but a non-null detections array'.format(
                            im['file']))
            else:
                if not isinstance(im['detections'],list):
                    raise ValueError('Invalid detections list for image {}'.format(im['file']))

            if is_video_file(im['file']) and (format_version >= 1.5):

                if 'frames_processed' not in im:
                    raise ValueError('Video without frames_processed field: {}'.format(im['file']))

            if is_video_file(im['file']) and (format_version >= 1.4):

                if 'frame_rate' not in im:
                    raise ValueError('Video without frame rate: {}'.format(im['file']))
                if im['frame_rate'] < 0:
                    if 'failure' not in im:
                        raise ValueError('Video with illegal frame rate {}: {}'.format(
                            str(im['frame_rate']),im['file']))
                if 'detections' in im and im['detections'] is not None:
                    for det in im['detections']:
                        if 'frame_number' not in det:
                            raise ValueError('Frame without frame number in video {}'.format(
                                im['file']))
                    frame_numbers = [det['frame_number'] for det in im['detections']] # noqa
                    # assert is_list_sorted(frame_numbers)

        # ...for each image


        ## Validation of other keys

        for k in d.keys():
            if (k not in typical_keys) and (k not in required_keys):
                validation_results['warnings'].append(
                    'Warning: non-standard key {} present at file level'.format(k))

    except Exception as e:

        if options.raise_errors:
            raise
        else:
            validation_results['errors'].append(str(e))

    # ...try/except

    if options.return_data:
        to_return = d
    else:
        to_return = {}

    to_return['validation_results'] = validation_results

    return to_return

# ...def validate_batch_results(...)


#%% Interactive driver(s)

if False:

    #%% Validate all .json files in the MD test suite

    from megadetector.utils.path_utils import recursive_file_list
    filenames = recursive_file_list(os.path.expanduser('~/AppData/Local/Temp/md-tests'))
    filenames = [fn for fn in filenames if fn.endswith('.json')]
    filenames = [fn for fn in filenames if 'detectionIndex' not in fn]

    options = ValidateBatchResultsOptions()
    options.check_image_existence = False
    options.relative_path_base = None # r'g:\temp\test-videos'

    for json_filename in filenames:
        results = validate_batch_results(json_filename,options)
        if len(results['validation_results']['warnings']) > 0:
            print('Warnings in file {}:'.format(json_filename))
            for s in results['validation_results']['warnings']:
                print(s)
            print('')
        assert len(results['validation_results']['errors']) == 0


#%% Command-line driver

def main(): # noqa

    options = ValidateBatchResultsOptions()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'json_filename',
        help='path to .json file containing MegaDetector results')
    parser.add_argument(
        '--check_image_existence', action='store_true',
        help='check that all images referred to in the results file exist')
    parser.add_argument(
        '--relative_path_base', default=None,
        help='if --check_image_existence is specified and paths are relative, use this as the base folder')
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    args_to_object(args, options)

    validate_batch_results(args.json_filename,options)


if __name__ == '__main__':
    main()



"""

create_crop_folder.py

Given a MegaDetector .json file and a folder of images, creates a new folder
of images representing all above-threshold crops from the original folder.

"""

#%% Constants and imports

import os
import json
import argparse

from tqdm import tqdm

from multiprocessing.pool import Pool, ThreadPool
from collections import defaultdict
from functools import partial

from megadetector.utils.path_utils import insert_before_extension
from megadetector.utils.ct_utils import invert_dictionary
from megadetector.utils.ct_utils import is_list_sorted
from megadetector.visualization.visualization_utils import crop_image
from megadetector.visualization.visualization_utils import exif_preserving_save


#%% Support classes

class CreateCropFolderOptions:
    """
    Options used to parameterize create_crop_folder().
    """

    def __init__(self):

        #: Confidence threshold determining which detections get written
        self.confidence_threshold = 0.1

        #: Number of pixels to expand each crop
        self.expansion = 0

        #: JPEG quality to use for saving crops (None for default)
        self.quality = 95

        #: Whether to overwrite existing images
        self.overwrite = True

        #: Number of concurrent workers
        self.n_workers = 8

        #: Whether to use processes ('process') or threads ('thread') for parallelization
        self.pool_type = 'thread'

        #: Include only these categories, or None to include all
        #:
        #: options.category_names_to_include = ['animal']
        self.category_names_to_include = None


#%% Support functions

def _get_crop_filename(image_fn,crop_id):
    """
    Generate crop filenames in a consistent way.
    """

    if isinstance(crop_id,int):
        crop_id = str(crop_id).zfill(3)
    assert isinstance(crop_id,str)
    return insert_before_extension(image_fn,'crop_' + crop_id)


def _generate_crops_for_single_image(crops_this_image,
                                     input_folder,
                                     output_folder,
                                     options):
    """
    Generate all the crops required for a single image.

    Args:
        crops_this_image (list of dict): list of dicts with at least keys
            'image_fn_relative', 'crop_id'
        input_folder (str): input folder (whole images)
        output_folder (crops): output folder (crops)
        options (CreateCropFolderOptions): cropping options
    """

    if len(crops_this_image) == 0:
        return

    image_fn_relative = crops_this_image[0]['image_fn_relative']
    input_fn_abs = os.path.join(input_folder,image_fn_relative)
    assert os.path.isfile(input_fn_abs)

    detections_to_crop = [c['detection'] for c in crops_this_image]

    cropped_images = crop_image(detections_to_crop,
                                input_fn_abs,
                                confidence_threshold=0,
                                expansion=options.expansion)

    assert len(cropped_images) == len(crops_this_image)

    # i_crop = 0; crop_info = crops_this_image[0]
    for i_crop,crop_info in enumerate(crops_this_image):

        assert crop_info['image_fn_relative'] == image_fn_relative
        crop_filename_relative = _get_crop_filename(image_fn_relative, crop_info['crop_id'])
        crop_filename_abs = os.path.join(output_folder,crop_filename_relative).replace('\\','/')

        if os.path.isfile(crop_filename_abs) and not options.overwrite:
            continue

        cropped_image = cropped_images[i_crop]
        os.makedirs(os.path.dirname(crop_filename_abs),exist_ok=True)
        exif_preserving_save(cropped_image,crop_filename_abs,quality=options.quality)

    # ...for each crop


#%% Main function

def crop_results_to_image_results(image_results_file_with_crop_ids,
                                  crop_results_file,
                                  output_file,
                                  delete_crop_information=True,
                                  require_identical_detection_categories=True,
                                  restrict_to_top_n=-1,
                                  crop_results_prefix=None,
                                  detections_without_classification_handling='error'):
    """
    This function is intended to be run after you have:

        1. Run MegaDetector on a folder
        2. Generated a crop folder using create_crop_folder
        3. Run a species classifier on those crops

    This function will take the crop-level results and transform them back
    to the original images.  Classification categories, if available, are taken
    from [crop_results_file].

    Args:
        image_results_file_with_crop_ids (str): results file for the original images,
            containing crop IDs, likely generated via create_crop_folder.  All
            non-standard fields in this file will be passed along to [output_file].
        crop_results_file (str): results file for the crop folder
        output_file (str): output .json file, containing crop-level classifications
            mapped back to the image level.
        delete_crop_information (bool, optional): whether to delete the "crop_id" and
            "crop_filename_relative" fields from each detection, if present.
        require_identical_detection_categories (bool, optional): if True, error if
            the image-level and crop-level detection categories are different.  If False,
            ignore the crop-level detection categories.
        restrict_to_top_n (int, optional): If >0, removes all but the top N classification
            results for each detection.
        crop_results_prefix (str, optional): if not None, removes this prefix from crop
            results filenames.  Intended to support the case where the crop results
            use absolute paths.
        detections_without_classification_handling (str, optional): what to do when we
            encounter a crop that doesn't appear in classification results: 'error',
            or 'include' ("include" means "leave the detection alone, without classifications"
    """

    ##%% Validate inputs

    assert os.path.isfile(image_results_file_with_crop_ids), \
        'Could not find image-level input file {}'.format(image_results_file_with_crop_ids)
    assert os.path.isfile(crop_results_file), \
        'Could not find crop results file {}'.format(crop_results_file)
    os.makedirs(os.path.dirname(output_file),exist_ok=True)


    ##%% Read input files

    print('Reading input...')

    with open(image_results_file_with_crop_ids,'r') as f:
        image_results_with_crop_ids = json.load(f)
    with open(crop_results_file,'r') as f:
        crop_results = json.load(f)

    # Find all the detection categories that need to be consistent
    used_detection_category_ids = set()
    for im in tqdm(image_results_with_crop_ids['images']):
        if 'detections' not in im or im['detections'] is None:
            continue
        for det in im['detections']:
            if 'crop_id' in det:
                used_detection_category_ids.add(det['category'])

    # Make sure the detection categories that matter are consistent across the two files
    if require_identical_detection_categories:
        for category_id in used_detection_category_ids:
            category_name = image_results_with_crop_ids['detection_categories'][category_id]
            assert category_id in crop_results['detection_categories'] and \
                category_name == crop_results['detection_categories'][category_id], \
                    'Crop results and detection results use incompatible categories'

    crop_filename_to_results = {}

    # im = crop_results['images'][0]
    for im in crop_results['images']:
        fn = im['file']
        # Possibly remove a prefix from each filename
        if (crop_results_prefix is not None) and (crop_results_prefix in fn):
            if fn.startswith(crop_results_prefix):
                fn = fn.replace(crop_results_prefix,'',1)
                im['file'] = fn
        crop_filename_to_results[fn] = im

    if 'classification_categories' in crop_results:
        image_results_with_crop_ids['classification_categories'] = \
            crop_results['classification_categories']

    if 'classification_category_descriptions' in crop_results:
        image_results_with_crop_ids['classification_category_descriptions'] = \
            crop_results['classification_category_descriptions']


    ##%% Read classifications from crop results, merge into image-level results

    print('Reading classification results...')

    n_skipped_detections = 0

    # Loop over the original image-level detections
    #
    # im = image_results_with_crop_ids['images'][0]
    for i_image,im in tqdm(enumerate(image_results_with_crop_ids['images']),
                           total=len(image_results_with_crop_ids['images'])):

        if 'detections' not in im or im['detections'] is None:
            continue

        # i_det = 0; det = im['detections'][i_det]
        for det in im['detections']:

            if 'classifications' in det:
                del det['classifications']

            if 'crop_id' in det:

                # We may be skipping detections with no classification results
                skip_detection = False

                # Find the corresponding crop in the classification results
                crop_filename_relative = det['crop_filename_relative']
                if crop_filename_relative not in crop_filename_to_results:
                    if detections_without_classification_handling == 'error':
                        raise ValueError('Crop lookup error: {}'.format(crop_filename_relative))
                    elif detections_without_classification_handling == 'include':
                        # Leave this detection unclassified
                        skip_detection = True
                    else:
                        raise ValueError(
                            'Illegal value for detections_without_classification_handling: {}'.format(
                                detections_without_classification_handling
                        ))

                if not skip_detection:

                    crop_results_this_detection = crop_filename_to_results[crop_filename_relative]

                    # Consistency checking
                    assert crop_results_this_detection['file'] == crop_filename_relative, \
                        'Crop filename mismatch'
                    assert len(crop_results_this_detection['detections']) == 1, \
                        'Multiple crop results for a single detection'
                    assert crop_results_this_detection['detections'][0]['bbox'] == [0,0,1,1], \
                        'Invalid crop bounding box'

                    # This check was helpful for the case where crop-level results had already
                    # taken detection confidence values from detector output by construct, but this isn't
                    # really meaningful for most cases.
                    # assert abs(crop_results_this_detection['detections'][0]['conf'] - det['conf']) < 0.01

                    if require_identical_detection_categories:
                        assert crop_results_this_detection['detections'][0]['category'] == det['category']

                    # Copy the crop-level classifications
                    det['classifications'] = crop_results_this_detection['detections'][0]['classifications']
                    confidence_values = [x[1] for x in det['classifications']]
                    assert is_list_sorted(confidence_values,reverse=True)
                    if restrict_to_top_n > 0:
                        det['classifications'] = det['classifications'][0:restrict_to_top_n]

            if delete_crop_information:
                if 'crop_id' in det:
                    del det['crop_id']
                if 'crop_filename_relative' in det:
                    del det['crop_filename_relative']

        # ...for each detection

    # ...for each image

    if n_skipped_detections > 0:
        print('Skipped {} detections'.format(n_skipped_detections))


    ##%% Write output file

    print('Writing output file...')

    with open(output_file,'w') as f:
        json.dump(image_results_with_crop_ids,f,indent=1)

# ...def crop_results_to_image_results(...)


def create_crop_folder(input_file,
                       input_folder,
                       output_folder,
                       output_file=None,
                       crops_output_file=None,
                       options=None):
    """
    Given a MegaDetector .json file and a folder of images, creates a new folder
    of images representing all above-threshold crops from the original folder.

    Optionally writes a new .json file that attaches unique IDs to each detection.

    Args:
        input_file (str): MD-formatted .json file to process
        input_folder (str): Input image folder
        output_folder (str): Output (cropped) image folder
        output_file (str, optional): new .json file that attaches unique IDs to each detection.
        crops_output_file (str, optional): new .json file that includes whole-image detections
            for each of the crops, using confidence values from the original results
        options (CreateCropFolderOptions, optional): crop parameters
    """

    ## Validate options, prepare output folders

    if options is None:
        options = CreateCropFolderOptions()

    assert os.path.isfile(input_file), 'Input file {} not found'.format(input_file)
    assert os.path.isdir(input_folder), 'Input folder {} not found'.format(input_folder)
    os.makedirs(output_folder,exist_ok=True)
    if output_file is not None:
        os.makedirs(os.path.dirname(output_file),exist_ok=True)


    ##%% Read input

    print('Reading MD results file...')
    with open(input_file,'r') as f:
        detection_results = json.load(f)

    category_ids_to_include = None

    if options.category_names_to_include is not None:
        category_id_to_name = detection_results['detection_categories']
        category_name_to_id = invert_dictionary(category_id_to_name)
        category_ids_to_include = set()
        for category_name in options.category_names_to_include:
            assert category_name in category_name_to_id, \
                'Unrecognized category name {}'.format(category_name)
            category_ids_to_include.add(category_name_to_id[category_name])

    ##%% Make a list of crops that we need to create

    # Maps input images to list of dicts, with keys 'crop_id','detection'
    image_fn_relative_to_crops = defaultdict(list)
    n_crops = 0

    n_detections_excluded_by_category = 0

    # im = detection_results['images'][0]
    for i_image,im in enumerate(detection_results['images']):

        if 'detections' not in im or im['detections'] is None or len(im['detections']) == 0:
            continue

        detections_this_image = im['detections']

        image_fn_relative = im['file']

        for i_detection,det in enumerate(detections_this_image):

            if det['conf'] < options.confidence_threshold:
                continue

            if (category_ids_to_include is not None) and \
                (det['category'] not in category_ids_to_include):
                n_detections_excluded_by_category += 1
                continue

            det['crop_id'] = i_detection

            crop_info = {'image_fn_relative':image_fn_relative,
                         'crop_id':i_detection,
                         'detection':det}

            crop_filename_relative = _get_crop_filename(image_fn_relative,
                                                        crop_info['crop_id'])
            det['crop_filename_relative'] = crop_filename_relative

            image_fn_relative_to_crops[image_fn_relative].append(crop_info)
            n_crops += 1

    # ...for each input image

    print('Prepared a list of {} crops from {} of {} input images'.format(
        n_crops,len(image_fn_relative_to_crops),len(detection_results['images'])))

    if n_detections_excluded_by_category > 0:
        print('Excluded {} detections by category'.format(n_detections_excluded_by_category))

    ##%% Generate crops

    if options.n_workers <= 1:

        # image_fn_relative = next(iter(image_fn_relative_to_crops))
        for image_fn_relative in tqdm(image_fn_relative_to_crops.keys()):
            crops_this_image = image_fn_relative_to_crops[image_fn_relative]
            _generate_crops_for_single_image(crops_this_image=crops_this_image,
                                             input_folder=input_folder,
                                             output_folder=output_folder,
                                             options=options)

    else:

        print('Creating a {} pool with {} workers'.format(options.pool_type,options.n_workers))
        pool = None
        try:
            if options.pool_type == 'thread':
                pool = ThreadPool(options.n_workers)
            else:
                assert options.pool_type == 'process'
                pool = Pool(options.n_workers)

            # Each element in this list is the list of crops for a single image
            crop_lists = list(image_fn_relative_to_crops.values())

            with tqdm(total=len(image_fn_relative_to_crops)) as pbar:
                for i,_ in enumerate(pool.imap_unordered(partial(
                            _generate_crops_for_single_image,
                                input_folder=input_folder,
                                output_folder=output_folder,
                                options=options),
                            crop_lists)):
                    pbar.update()
        finally:
            if pool is not None:
                pool.close()
                pool.join()
                print("Pool closed and joined for crop folder creation")

    # ...if we're using parallel processing


    ##%% Write output file

    if output_file is not None:
        with open(output_file,'w') as f:
            json.dump(detection_results,f,indent=1)

    if crops_output_file is not None:

        original_images = detection_results['images']

        detection_results_cropped = detection_results
        detection_results_cropped['images'] = []

        # im = original_images[0]
        for im in original_images:

            if 'detections' not in im or im['detections'] is None or len(im['detections']) == 0:
                continue

            detections_this_image = im['detections']
            image_fn_relative = im['file']

            for i_detection,det in enumerate(detections_this_image):

                if 'crop_id' in det:
                    im_out = {}
                    im_out['file'] = det['crop_filename_relative']
                    det_out = {}
                    det_out['category'] = det['category']
                    det_out['conf'] = det['conf']
                    det_out['bbox'] = [0, 0, 1, 1]
                    im_out['detections'] = [det_out]
                    detection_results_cropped['images'].append(im_out)

                # ...if we need to include this crop in the new .json file

            # ...for each crop

        # ...for each original image

        with open(crops_output_file,'w') as f:
            json.dump(detection_results_cropped,f,indent=1)

# ...def create_crop_folder()


#%% Command-line driver

def main():
    """
    Command-line interface for creating a crop folder from MegaDetector results.
    """

    parser = argparse.ArgumentParser(
        description='Create a folder of crops from MegaDetector results'
    )
    parser.add_argument(
        'input_file',
        type=str,
        help='Path to the MegaDetector .json results file'
    )
    parser.add_argument(
        'input_folder',
        type=str,
        help='Path to the folder containing the original images'
    )
    parser.add_argument(
        'output_folder',
        type=str,
        help='Path to the folder where cropped images will be saved'
    )
    parser.add_argument(
        '--output_file',
        type=str,
        default=None,
        help='Path to save the modified MegaDetector .json file (with crop IDs and filenames)'
    )
    parser.add_argument(
        '--crops_output_file',
        type=str,
        default=None,
        help='Path to save a new .json file for the crops themselves (with full-image detections for each crop)'
    )
    parser.add_argument(
        '--confidence_threshold',
        type=float,
        default=0.1,
        help='Confidence threshold for detections to be cropped (default: 0.1)'
    )
    parser.add_argument(
        '--expansion',
        type=int,
        default=0,
        help='Number of pixels to expand each crop (default: 0)'
    )
    parser.add_argument(
        '--quality',
        type=int,
        default=95,
        help='JPEG quality for saving crops (default: 95)'
    )
    parser.add_argument(
        '--overwrite',
        type=str,
        default='true',
        choices=['true', 'false'],
        help="Overwrite existing crop images (default: 'true')"
    )
    parser.add_argument(
        '--n_workers',
        type=int,
        default=8,
        help='Number of concurrent workers (default: 8)'
    )
    parser.add_argument(
        '--pool_type',
        type=str,
        default='thread',
        choices=['thread', 'process'],
        help="Type of parallelism to use ('thread' or 'process', default: 'thread')"
    )
    parser.add_argument(
        '--category_names',
        type=str,
        default=None,
        help="Comma-separated list of category names to include " + \
             "(e.g., 'animal,person'). If None (default), all categories are included."
    )

    args = parser.parse_args()

    options = CreateCropFolderOptions()
    options.confidence_threshold = args.confidence_threshold
    options.expansion = args.expansion
    options.quality = args.quality
    options.overwrite = (args.overwrite.lower() == 'true')
    options.n_workers = args.n_workers
    options.pool_type = args.pool_type

    if args.category_names:
        options.category_names_to_include = [name.strip() for name in args.category_names.split(',')]
    else:
        options.category_names_to_include = None

    print('Starting crop folder creation...')
    print('Input MD results: {}'.format(args.input_file))
    print('Input image folder {}'.format(args.input_folder))
    print('Output crop folder: {}'.format(args.output_folder))

    if args.output_file:
        print('Modified MD results will be saved to {}'.format(args.output_file))
    if args.crops_output_file:
        print('Crops .json output will be saved to {}'.format(args.crops_output_file))

    create_crop_folder(
        input_file=args.input_file,
        input_folder=args.input_folder,
        output_folder=args.output_folder,
        output_file=args.output_file,
        crops_output_file=args.crops_output_file,
        options=options
    )

if __name__ == '__main__':
    main()



"""

train_classifier.py

Train a EfficientNet or ResNet classifier.

Currently the implementation of multi-label multi-class classification is
non-functional.

During training, start tensorboard from within the classification/ directory:
  tensorboard --logdir run --bind_all --samples_per_plugin scalars=0,images=0

"""

#%% Imports and constants

from __future__ import annotations

import json
import os
import argparse

from collections import defaultdict
from collections.abc import Callable, Mapping, MutableMapping, Sequence
from datetime import datetime
from typing import Any

import numpy as np
import PIL.Image
import sklearn.metrics
import tqdm

import torch
from torch.utils import tensorboard
import torchvision as tv
from torchvision.datasets.folder import default_loader

from megadetector.classification import efficientnet, evaluate_model
from megadetector.classification.train_utils import (
    HeapItem, recall_from_confusion_matrix, add_to_heap, fig_to_img,
    imgs_with_confidences, load_dataset_csv, prefix_all_keys)
from megadetector.visualization import plot_utils


#%% Example usage

"""
    python train_classifier.py run_idfg /ssd/crops_sq \
        -m "efficientnet-b0" --pretrained --finetune --label-weighted \
        --epochs 50 --batch-size 512 --lr 1e-4 \
        --num-workers 12 --seed 123 \
        --logdir run_idfg
"""

# mean/std values from https://pytorch.org/docs/stable/torchvision/models.html
MEANS = np.asarray([0.485, 0.456, 0.406])
STDS = np.asarray([0.229, 0.224, 0.225])

VALID_MODELS = sorted(
    set(efficientnet.VALID_MODELS) |
    {'resnet101', 'resnet152', 'resnet18', 'resnet34', 'resnet50'})


class AverageMeter:
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()

    def reset(self) -> None:
        self.val = 0.0
        self.avg = 0.0
        self.sum = 0.0
        self.count = 0

    def update(self, val: float, n: int = 1) -> None:
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


class SimpleDataset(torch.utils.data.Dataset):
    """A simple dataset that simply returns images and labels."""

    def __init__(self,
                 img_files: Sequence[str],
                 labels: Sequence[Any],
                 sample_weights: Sequence[float] | None = None,
                 img_base_dir: str = '',
                 transform: Callable[[PIL.Image.Image], Any] | None = None,
                 target_transform: Callable[[Any], Any] | None = None):
        """Creates a SimpleDataset."""
        self.img_files = img_files
        self.labels = labels
        self.sample_weights = sample_weights
        self.img_base_dir = img_base_dir
        self.transform = transform
        self.target_transform = target_transform

        self.len = len(img_files)
        assert len(labels) == self.len
        if sample_weights is not None:
            assert len(sample_weights) == self.len

    def __getitem__(self, index: int) -> tuple[Any, ...]:
        """
        Args:
            index: int

        Returns: tuple, (sample, target) or (sample, target, sample_weight)
        """
        img_file = self.img_files[index]
        img = default_loader(os.path.join(self.img_base_dir, img_file))
        if self.transform is not None:
            img = self.transform(img)
        target = self.labels[index]
        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.sample_weights is not None:
            return img, target, img_file, self.sample_weights[index]
        return img, target, img_file

    def __len__(self) -> int:
        return self.len


def create_dataloaders(
        dataset_csv_path: str,
        label_index_json_path: str,
        splits_json_path: str,
        cropped_images_dir: str,
        img_size: int,
        multilabel: bool,
        label_weighted: bool,
        weight_by_detection_conf: bool | str,
        batch_size: int,
        num_workers: int,
        augment_train: bool
        ) -> tuple[dict[str, torch.utils.data.DataLoader], list[str]]:
    """
    Args:
        dataset_csv_path: str, path to CSV file with columns
            ['dataset', 'location', 'label'], where label is a comma-delimited
            list of labels
        splits_json_path: str, path to JSON file
        augment_train: bool, whether to shuffle/augment the training set

    Returns:
        datasets: dict, maps split to DataLoader
        label_names: list of str, label names in order of label id
    """
    df, label_names, split_to_locs = load_dataset_csv(
        dataset_csv_path, label_index_json_path, splits_json_path,
        multilabel=multilabel, label_weighted=label_weighted,
        weight_by_detection_conf=weight_by_detection_conf)

    # define the transforms
    normalize = tv.transforms.Normalize(mean=MEANS, std=STDS, inplace=True)
    train_transform = tv.transforms.Compose([
        tv.transforms.RandomResizedCrop(img_size),
        tv.transforms.RandomRotation(degrees=(-90, 90)),
        tv.transforms.RandomHorizontalFlip(p=0.5),
        tv.transforms.RandomVerticalFlip(p=0.1),
        tv.transforms.RandomGrayscale(p=0.1),
        tv.transforms.ColorJitter(brightness=.25, contrast=.25, saturation=.25),
        tv.transforms.ToTensor(),
        normalize
    ])
    test_transform = tv.transforms.Compose([
        # resizes smaller edge to img_size
        tv.transforms.Resize(img_size, interpolation=PIL.Image.BICUBIC),
        tv.transforms.CenterCrop(img_size),
        tv.transforms.ToTensor(),
        normalize
    ])

    dataloaders = {}
    for split, locs in split_to_locs.items():
        is_train = (split == 'train') and augment_train
        split_df = df[df['dataset_location'].isin(locs)]

        sampler: torch.utils.data.Sampler | None = None
        weights = None
        if label_weighted or weight_by_detection_conf:
            # weights sums to:
            # - if weight_by_detection_conf: (# images in split - conf delta)
            # - otherwise: # images in split
            weights = split_df['weights'].to_numpy()
            if not weight_by_detection_conf:
                assert np.isclose(weights.sum(), len(split_df))
            if is_train:
                sampler = torch.utils.data.WeightedRandomSampler(
                    weights, num_samples=len(split_df), replacement=True)
        elif is_train:
            # for normal (non-weighted) shuffling
            sampler = torch.utils.data.SubsetRandomSampler(range(len(split_df)))

        dataset = SimpleDataset(
            img_files=split_df['path'].tolist(),
            labels=split_df['label_index'].tolist(),
            sample_weights=weights,
            img_base_dir=cropped_images_dir,
            transform=train_transform if is_train else test_transform)
        assert len(dataset) > 0
        dataloaders[split] = torch.utils.data.DataLoader(
            dataset, batch_size=batch_size, sampler=sampler,
            num_workers=num_workers, pin_memory=True)

    return dataloaders, label_names


def set_finetune(model: torch.nn.Module, model_name: str, finetune: bool
                 ) -> None:
    """Set the 'requires_grad' on each model parameter according to whether or
    not we are fine-tuning the model.
    """
    if finetune:
        if 'efficientnet' in model_name:
            final_layer = model._fc  # pylint: disable=protected-access
        else:  # torchvision resnet
            final_layer = model.fc
        assert isinstance(final_layer, torch.nn.Module)

        # set all parameters to not require gradients except final FC layer
        model.requires_grad_(False)
        for param in final_layer.parameters():
            param.requires_grad = True
    else:
        model.requires_grad_(True)


def build_model(model_name: str, num_classes: int, pretrained: bool | str,
                finetune: bool) -> torch.nn.Module:
    """Creates a model with an EfficientNet or ResNet base. The model outputs
    unnormalized logits.

    Args:
        model_name: str, name of EfficientNet or Resnet model
        num_classes: int, number of classes for output layer
        pretrained: bool or str, (bool) whether to initialize to ImageNet
            weights, (str) path to checkpoint
        finetune: bool, whether to freeze all layers except the final FC layer

    Returns: torch.nn.Module, model loaded on CPU
    """
    assert model_name in VALID_MODELS

    if 'efficientnet' in model_name:
        if pretrained is True:
            model = efficientnet.EfficientNet.from_pretrained(
                model_name, num_classes=num_classes)
        else:
            model = efficientnet.EfficientNet.from_name(
                model_name, num_classes=num_classes)
    else:
        model_class = getattr(tv.models, model_name)
        model = model_class(pretrained=(pretrained is True))

        # replace final fully-connected layer (which has 1000 ImageNet classes)
        model.fc = torch.nn.Linear(model.fc.in_features, num_classes)

    if isinstance(pretrained, str):
        print(f'Loading saved weights from {pretrained}')
        ckpt = torch.load(pretrained, map_location='cpu')
        model.load_state_dict(ckpt['model'])

    assert all(p.requires_grad for p in model.parameters())
    set_finetune(model=model, model_name=model_name, finetune=finetune)
    return model


def prep_device(model: torch.nn.Module, device_id: int | None = None
                ) -> tuple[torch.nn.Module, torch.device]:
    """Place model on appropriate device.

    Args:
        model: torch.nn.Module, not already wrapped with DataParallel
        device_id: optional int, GPU device to use
            if None, then uses DataParallel when possible
            if specified, then only uses specified device

    Returns:
        model: torch.nn.Module, model placed on <device>, wrapped with
            DataParallel if more than 1 GPU is found
        device: torch.device, 'cuda:{device_id}' if GPU is found, otherwise 'cpu'
    """
    # detect GPU, use all if available
    if torch.cuda.is_available():
        print('CUDA available')
        torch.backends.cudnn.benchmark = True
        if device_id is not None:
            print(f'Starting CUDA device {device_id}')
            device = torch.device(f'cuda:{device_id}')
        else:
            device = torch.device('cuda:0')
            device_ids = list(range(torch.cuda.device_count()))
            if len(device_ids) > 1:
                print(f'Found multiple devices, enabling data parallelism ({device_ids})')
                model = torch.nn.DataParallel(model, device_ids=device_ids)
    else:
        print('CUDA not available, running on the CPU')
        device = torch.device('cpu')
    model.to(device)  # in-place
    return model, device


def main(dataset_dir: str,
         cropped_images_dir: str,
         multilabel: bool,
         model_name: str,
         pretrained: bool | str,
         finetune: int,
         label_weighted: bool,
         weight_by_detection_conf: bool | str,
         epochs: int,
         batch_size: int,
         lr: float,
         weight_decay: float,
         num_workers: int,
         logdir: str,
         log_extreme_examples: int,
         seed: int | None = None) -> None:
    """Main function."""
    # input validation
    assert os.path.exists(dataset_dir)
    assert os.path.exists(cropped_images_dir)
    if isinstance(weight_by_detection_conf, str):
        assert os.path.exists(weight_by_detection_conf)
    if isinstance(pretrained, str):
        assert os.path.exists(pretrained)

    # set seed
    seed = np.random.randint(10_000) if seed is None else seed
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    # create logdir and save params
    params = dict(locals())  # make a copy
    timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')  # '20200722_110816'
    logdir = os.path.join(logdir, timestamp)
    os.makedirs(logdir, exist_ok=True)
    print('Created logdir:', logdir)
    params_json_path = os.path.join(logdir, 'params.json')
    with open(params_json_path, 'w') as f:
        json.dump(params, f, indent=1)

    if 'efficientnet' in model_name:
        img_size = efficientnet.EfficientNet.get_image_size(model_name)
    else:
        img_size = 224

    # create dataloaders and log the index_to_label mapping
    print('Creating dataloaders')
    loaders, label_names = create_dataloaders(
        dataset_csv_path=os.path.join(dataset_dir, 'classification_ds.csv'),
        label_index_json_path=os.path.join(dataset_dir, 'label_index.json'),
        splits_json_path=os.path.join(dataset_dir, 'splits.json'),
        cropped_images_dir=cropped_images_dir,
        img_size=img_size,
        multilabel=multilabel,
        label_weighted=label_weighted,
        weight_by_detection_conf=weight_by_detection_conf,
        batch_size=batch_size,
        num_workers=num_workers,
        augment_train=True)

    writer = tensorboard.SummaryWriter(logdir)

    # create model
    model = build_model(model_name, num_classes=len(label_names),
                        pretrained=pretrained, finetune=finetune > 0)
    model, device = prep_device(model)

    # define loss function and optimizer
    loss_fn: torch.nn.Module
    if multilabel:
        loss_fn = torch.nn.BCEWithLogitsLoss(reduction='none').to(device)
    else:
        loss_fn = torch.nn.CrossEntropyLoss(reduction='none').to(device)

    # using EfficientNet training defaults
    # - batch norm momentum: 0.99
    # - optimizer: RMSProp, decay 0.9 and momentum 0.9
    # - epochs: 350
    # - learning rate: 0.256, decays by 0.97 every 2.4 epochs
    # - weight decay: 1e-5
    optimizer: torch.optim.Optimizer
    if 'efficientnet' in model_name:
        optimizer = torch.optim.RMSprop(model.parameters(), lr, alpha=0.9,
                                        momentum=0.9, weight_decay=weight_decay)
        lr_scheduler = torch.optim.lr_scheduler.StepLR(
            optimizer=optimizer, step_size=1, gamma=0.97 ** (1 / 2.4))
    else:  # resnet
        optimizer = torch.optim.SGD(model.parameters(), lr, momentum=0.9,
                                    weight_decay=weight_decay)
        lr_scheduler = torch.optim.lr_scheduler.StepLR(
            optimizer=optimizer, step_size=8, gamma=0.1)  # lower every 8 epochs

    best_epoch_metrics: dict[str, float] = {}
    for epoch in range(epochs):
        print(f'Epoch: {epoch}')
        writer.add_scalar('lr', lr_scheduler.get_last_lr()[0], epoch)

        if epoch > 0 and finetune == epoch:
            print('Turning off fine-tune!')
            set_finetune(model, model_name, finetune=False)

        print('- train:')
        train_metrics, train_heaps, train_cm = run_epoch(
            model, loader=loaders['train'], weighted=False, device=device,
            loss_fn=loss_fn, finetune=finetune > epoch, optimizer=optimizer,
            k_extreme=log_extreme_examples)
        train_metrics = prefix_all_keys(train_metrics, prefix='train/')
        log_run('train', epoch, writer, label_names,
                metrics=train_metrics, heaps=train_heaps, cm=train_cm)
        del train_heaps

        print('- val:')
        val_metrics, val_heaps, val_cm = run_epoch(
            model, loader=loaders['val'], weighted=label_weighted,
            device=device, loss_fn=loss_fn, k_extreme=log_extreme_examples)
        val_metrics = prefix_all_keys(val_metrics, prefix='val/')
        log_run('val', epoch, writer, label_names,
                metrics=val_metrics, heaps=val_heaps, cm=val_cm)
        del val_heaps

        lr_scheduler.step()  # decrease the learning rate

        if val_metrics['val/acc_top1'] > best_epoch_metrics.get('val/acc_top1', 0):  # pylint: disable=line-too-long
            filename = os.path.join(logdir, f'ckpt_{epoch}.pt')
            print(f'New best model! Saving checkpoint to {filename}')
            state = {
                'epoch': epoch,
                'model': getattr(model, 'module', model).state_dict(),
                'val/acc': val_metrics['val/acc_top1'],
                'optimizer': optimizer.state_dict()
            }
            torch.save(state, filename)
            best_epoch_metrics.update(train_metrics)
            best_epoch_metrics.update(val_metrics)
            best_epoch_metrics['epoch'] = epoch

            print('- test:')
            test_metrics, test_heaps, test_cm = run_epoch(
                model, loader=loaders['test'], weighted=label_weighted,
                device=device, loss_fn=loss_fn, k_extreme=log_extreme_examples)
            test_metrics = prefix_all_keys(test_metrics, prefix='test/')
            log_run('test', epoch, writer, label_names,
                    metrics=test_metrics, heaps=test_heaps, cm=test_cm)
            del test_heaps

        # stop training after 8 epochs without improvement
        if epoch >= best_epoch_metrics['epoch'] + 8:
            break

    hparams_dict = {
        'model_name': model_name,
        'multilabel': multilabel,
        'finetune': finetune,
        'batch_size': batch_size,
        'epochs': epochs
    }
    metric_dict = prefix_all_keys(best_epoch_metrics, prefix='hparam/')
    writer.add_hparams(hparam_dict=hparams_dict, metric_dict=metric_dict)
    writer.close()

    # do a complete evaluation run
    best_epoch = best_epoch_metrics['epoch']
    evaluate_model.main(
        params_json_path=params_json_path,
        ckpt_path=os.path.join(logdir, f'ckpt_{best_epoch}.pt'),
        output_dir=logdir, splits=evaluate_model.SPLITS)


def log_run(split: str, epoch: int, writer: tensorboard.SummaryWriter,
            label_names: Sequence[str], metrics: MutableMapping[str, float],
            heaps: Mapping[str, Mapping[int, list[HeapItem]]] | None,
            cm: np.ndarray) -> None:
    """Logs the outputs (metrics, confusion matrix, tp/fp/fn images) from a
    single epoch run to Tensorboard.

    Args:
        metrics: dict, keys already prefixed with {split}/
    """
    per_label_recall = recall_from_confusion_matrix(cm, label_names)
    metrics.update(prefix_all_keys(per_label_recall, f'{split}/label_recall/'))

    # log metrics
    for metric, value in metrics.items():
        writer.add_scalar(metric, value, epoch)

    # log confusion matrix
    cm_fig = plot_utils.plot_confusion_matrix(cm, classes=label_names,
                                              normalize=True)
    cm_fig_img = fig_to_img(cm_fig)
    writer.add_image(tag=f'confusion_matrix/{split}', img_tensor=cm_fig_img,
                     global_step=epoch, dataformats='HWC')

    # log tp/fp/fn images
    if heaps is not None:
        for heap_type, heap_dict in heaps.items():
            log_images_with_confidence(writer, heap_dict, label_names,
                                       epoch=epoch, tag=f'{split}/{heap_type}')
    writer.flush()


def log_images_with_confidence(
        writer: tensorboard.SummaryWriter,
        heap_dict: Mapping[int, list[HeapItem]],
        label_names: Sequence[str],
        epoch: int,
        tag: str) -> None:
    """
    Note: performs image normalization in-place

    Args:
        writer: tensorboard.SummaryWriter
        heap_dict: dict, maps label_id to list of HeapItem, where each HeapItem
            data is a tuple (img, target, top3_conf, top3_preds, img_file)
        label_names: list of str, label names in order of label id
        epoch: int
        tag: str
    """
    # for every image: undo normalization, clamp to [0, 1], CHW -> HWC
    # - cannot be in-place, because the HeapItem might be in multiple heaps
    unnormalize = tv.transforms.Normalize(mean=-MEANS/STDS, std=1.0/STDS)
    for label_id, heap in heap_dict.items():
        label_name = label_names[label_id]

        imgs_list = []
        for item in sorted(heap, reverse=True):  # sort largest to smallest
            img = item.data[0].float()  # clamp() only supports fp32 on CPU
            img = unnormalize(img).clamp_(0, 1).permute(1, 2, 0)
            imgs_list.append((img, *item.data[1:]))

        fig, img_files = imgs_with_confidences(imgs_list, label_names)

        # writer.add_figure() has issues => using add_image() instead
        # writer.add_figure(f'{label_name}/{tag}', fig, global_step=epoch)
        writer.add_image(f'{label_name}/{tag}', fig_to_img(fig),
                         global_step=epoch, dataformats='HWC')
        writer.add_text(f'{label_name}/{tag}_files', '\n\n'.join(img_files),
                        global_step=epoch)


def track_extreme_examples(tp_heaps: dict[int, list[HeapItem]],
                           fp_heaps: dict[int, list[HeapItem]],
                           fn_heaps: dict[int, list[HeapItem]],
                           inputs: torch.Tensor,
                           labels: torch.Tensor,
                           img_files: Sequence[str],
                           logits: torch.Tensor,
                           k: int = 5) -> None:
    """Updates the k most extreme true-positive (tp), false-positive (fp), and
    false-negative (fn) examples with examples from this batch.

    Each HeapItem's data attribute is a tuple of:
    - img: torch.Tensor, shape [3, H, W], type float16, values in [0, 1]
    - label: int
    - top3_conf: list of float
    - top3_preds: list of float
    - img_file: str

    Args:
        *_heaps: dict, maps label_id (int) to heap of HeapItems
        inputs: torch.Tensor, shape [batch_size, 3, H, W]
        labels: torch.Tensor, shape [batch_size]
        img_files: list of str
        logits: torch.Tensor, shape [batch_size, num_classes]
        k: int, number of examples to track
    """
    with torch.no_grad():
        inputs = inputs.detach().to(device='cpu', dtype=torch.float16)
        labels_list = labels.tolist()
        batch_probs = torch.nn.functional.softmax(logits, dim=1).cpu()
        zipped = zip(inputs, labels_list, batch_probs, img_files)  # all on CPU
        for img, label, confs, img_file in zipped:
            label_conf = confs[label].item()

            top3_conf, top3_preds = confs.topk(3)
            top3_conf = top3_conf.tolist()
            top3_preds = top3_preds.tolist()

            data = [img, label, top3_conf, top3_preds, img_file]
            if top3_preds[0] == label:  # true positive
                item = HeapItem(priority=label_conf - top3_conf[1], data=data)
                add_to_heap(tp_heaps[label], item, k=k)
            else:
                # false positive for top3_pred[0]
                # false negative for label
                item = HeapItem(priority=top3_conf[0] - label_conf, data=data)
                add_to_heap(fp_heaps[top3_preds[0]], item, k=k)
                add_to_heap(fn_heaps[label], item, k=k)


def correct(outputs: torch.Tensor, labels: torch.Tensor,
            weights: torch.Tensor | None = None,
            top: Sequence[int] = (1,)) -> dict[int, float]:
    """
    Args:
        outputs: torch.Tensor, shape [N, num_classes],
            either logits (pre-softmax) or probabilities
        labels: torch.Tensor, shape [N]
        weights: optional torch.Tensor, shape [N]
        top: tuple of int, list of values of k for calculating top-K accuracy

    Returns: dict, maps k to (weighted) # of correct predictions @ each k
    """
    with torch.no_grad():
        # preds and labels both have shape [N, k]
        _, preds = outputs.topk(k=max(top), dim=1, largest=True, sorted=True)
        labels = labels.view(-1, 1).expand_as(preds)

        corrects = preds.eq(labels).cumsum(dim=1)  # shape [N, k]
        if weights is None:
            corrects = corrects.sum(dim=0)  # shape [k]
        else:
            corrects = weights.matmul(corrects.to(weights.dtype))  # shape [k]
        tops = {k: corrects[k - 1].item() for k in top}
    return tops


def run_epoch(model: torch.nn.Module,
              loader: torch.utils.data.DataLoader,
              weighted: bool,
              device: torch.device,
              top: Sequence[int] = (1, 3),
              loss_fn: torch.nn.Module | None = None,
              finetune: bool = False,
              optimizer: torch.optim.Optimizer | None = None,
              k_extreme: int = 0
              ) -> tuple[
                  dict[str, float],
                  dict[str, dict[int, list[HeapItem]]] | None,
                  np.ndarray
              ]:
    """Runs for 1 epoch.

    Args:
        model: torch.nn.Module
        loader: torch.utils.data.DataLoader
        weighted: bool, whether to use sample weights in calculating loss and
            accuracy
        device: torch.device
        top: tuple of int, list of values of k for calculating top-K accuracy
        loss_fn: optional loss function, calculates per-example loss
        finetune: bool, if true sets model's dropout and BN layers to eval mode
        optimizer: optional optimizer
        k_extreme: int, # of tp/fp/fn examples to track for each label

    Returns:
        metrics: dict, metrics from epoch, contains keys:
            'loss': float, mean per-example loss over entire epoch,
                only included if loss_fn is not None
            'acc_top{k}': float, accuracy@k over the entire epoch
        heaps: dict, keys are ['tp', 'fp', 'fn'], values are heap_dicts,
            each heap_dict maps label_id (int) to a heap of <= 5 HeapItems with
            data attribute (img, target, top3_conf, top3_preds, img_file)
            - 'tp': priority is the difference between target confidence and
                2nd highest confidence
            - 'fp': priority is the difference between highest confidence and
                target confidence
            - 'fn': same as 'fp'
        confusion_matrix: np.ndarray, shape [num_classes, num_classes],
            C[i, j] = # of samples with true label i, predicted as label j
    """
    if optimizer is not None:
        assert loss_fn is not None

    # if evaluating or finetuning, set dropout and BN layers to eval mode
    model.train(optimizer is not None and not finetune)

    if loss_fn is not None:
        losses = AverageMeter()
    accuracies_topk = {k: AverageMeter() for k in top}  # acc@k

    # for each label, track k_extreme most-confident and least-confident images
    if k_extreme > 0:
        tp_heaps: dict[int, list[HeapItem]] = defaultdict(list)
        fp_heaps: dict[int, list[HeapItem]] = defaultdict(list)
        fn_heaps: dict[int, list[HeapItem]] = defaultdict(list)

    all_labels = np.zeros(len(loader.dataset), dtype=np.int32)
    all_preds = np.zeros_like(all_labels)
    end_i = 0

    tqdm_loader = tqdm.tqdm(loader)
    with torch.set_grad_enabled(optimizer is not None):
        for batch in tqdm_loader:
            if weighted:
                inputs, labels, img_files, weights = batch
                weights = weights.to(device, non_blocking=True)
            else:
                # even if batch contains sample weights, don't use them
                inputs, labels, img_files = batch[0:3]
                weights = None

            inputs = inputs.to(device, non_blocking=True)

            batch_size = labels.size(0)
            start_i = end_i
            end_i = start_i + batch_size
            all_labels[start_i:end_i] = labels

            desc = []
            labels = labels.to(device, non_blocking=True)
            outputs = model(inputs)
            all_preds[start_i:end_i] = outputs.detach().argmax(dim=1).cpu()

            if loss_fn is not None:
                loss = loss_fn(outputs, labels)
                if weights is not None:
                    loss *= weights
                loss = loss.mean()
                losses.update(loss.item(), n=batch_size)
                desc.append(f'Loss {losses.val:.4f} ({losses.avg:.4f})')
            if optimizer is not None:
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            top_correct = correct(outputs, labels, weights=weights, top=top)
            for k, acc in accuracies_topk.items():
                acc.update(top_correct[k] * (100. / batch_size), n=batch_size)
                desc.append(f'Acc@{k} {acc.val:.3f} ({acc.avg:.3f})')
            tqdm_loader.set_description(' '.join(desc))

            if k_extreme > 0:
                track_extreme_examples(tp_heaps, fp_heaps, fn_heaps, inputs,
                                       labels, img_files, outputs, k=k_extreme)

    num_classes = outputs.size(1)
    confusion_matrix = sklearn.metrics.confusion_matrix(
        all_labels, all_preds, labels=np.arange(num_classes))

    metrics = {}
    if loss_fn is not None:
        metrics['loss'] = losses.avg
    for k, acc in accuracies_topk.items():
        metrics[f'acc_top{k}'] = acc.avg
    heaps = None
    if k_extreme > 0:
        heaps = {'tp': tp_heaps, 'fp': fp_heaps, 'fn': fn_heaps}
    return metrics, heaps, confusion_matrix


def _parse_args() -> argparse.Namespace:
    """Parses arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Trains classifier.')
    parser.add_argument(
        'dataset_dir',
        help='path to directory containing: 1) classification dataset CSV, '
             '2) label index JSON, 3) splits JSON')
    parser.add_argument(
        'cropped_images_dir',
        help='path to local directory where image crops are saved')
    parser.add_argument(
        '--multilabel', action='store_true',
        help='for multi-label, multi-class classification')
    parser.add_argument(
        '-m', '--model-name', default='efficientnet-b0',
        choices=VALID_MODELS,
        help='which EfficientNet or Resnet model')
    parser.add_argument(
        '--pretrained', nargs='?', const=True, default=False,
        help='start with ImageNet pretrained model or a specific checkpoint')
    parser.add_argument(
        '--finetune', type=int, default=0,
        help='only fine tune the final fully-connected layer for the first '
             '<finetune> epochs')
    parser.add_argument(
        '--label-weighted', action='store_true',
        help='weight training samples to balance labels')
    parser.add_argument(
        '--weight-by-detection-conf', nargs='?', const=True, default=False,
        help='weight training examples by detection confidence. '
             'Optionally takes a .npz file for isotonic calibration.')
    parser.add_argument(
        '--epochs', type=int, default=0,
        help='number of epochs for training, 0 for eval-only')
    parser.add_argument(
        '--batch-size', type=int, default=256,
        help='batch size for both training and eval')
    parser.add_argument(
        '--lr', type=float,
        help='initial learning rate, defaults to (0.016 * batch_size / 256)')
    parser.add_argument(
        '--weight-decay', type=float, default=1e-5,
        help='weight decay')
    parser.add_argument(
        '--num-workers', type=int, default=8,
        help='# of workers for data loading')
    parser.add_argument(
        '--logdir', default='.',
        help='directory where TensorBoard logs and a params file are saved')
    parser.add_argument(
        '--log-extreme-examples', type=int, default=0,
        help='# of tp/fp/fn examples to log for each label and split per epoch')
    parser.add_argument(
        '--seed', type=int,
        help='random seed')
    return parser.parse_args()


if __name__ == '__main__':
    args = _parse_args()
    if args.lr is None:
        args.lr = 0.016 * args.batch_size / 256  # based on TF models repo
    main(dataset_dir=args.dataset_dir,
         cropped_images_dir=args.cropped_images_dir,
         multilabel=args.multilabel,
         model_name=args.model_name,
         pretrained=args.pretrained,
         finetune=args.finetune,
         label_weighted=args.label_weighted,
         weight_by_detection_conf=args.weight_by_detection_conf,
         epochs=args.epochs,
         batch_size=args.batch_size,
         lr=args.lr,
         weight_decay=args.weight_decay,
         num_workers=args.num_workers,
         logdir=args.logdir,
         log_extreme_examples=args.log_extreme_examples,
         seed=args.seed)



"""

url_utils.py

Frequently-used functions for downloading or manipulating URLs

"""

#%% Imports and constants

import os
import re
import urllib
import urllib.request
import urllib.error
import requests
import shutil
import pytest

from functools import partial
from tqdm import tqdm
from urllib.parse import urlparse
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool

from megadetector.utils.ct_utils import make_test_folder
from megadetector.utils.ct_utils import make_temp_folder

max_path_len = 255


#%% Download functions

class DownloadProgressBar:
    """
    Progress updater based on the progressbar2 package.

    https://stackoverflow.com/questions/37748105/how-to-use-progressbar-module-with-urlretrieve
    """


    def __init__(self):

        self.pbar = None


    def __call__(self, block_num, block_size, total_size): # noqa

        if not self.pbar:
            try:
                import progressbar # type: ignore
                self.pbar = progressbar.ProgressBar(max_value=total_size)
                self.pbar.start()
            except ImportError:
                self.pbar = None
                # print("ProgressBar not available, install 'progressbar2' for visual progress.")

        if self.pbar:
            downloaded = block_num * block_size
            if downloaded < total_size:
                self.pbar.update(downloaded)
            else:
                self.pbar.finish()


def download_url(url,
                 destination_filename=None,
                 progress_updater=None,
                 force_download=False,
                 verbose=True,
                 escape_spaces=True):
    """
    Downloads a URL to a file.  If no file is specified, creates a temporary file,
    making a best effort to avoid filename collisions.

    Prints some diagnostic information and makes sure to omit SAS tokens from printouts.

    Args:
        url (str): the URL to download
        destination_filename (str, optional): the target filename; if None, will create
            a file in system temp space
        progress_updater (object or bool, optional): can be "None", "False", "True", or a
            specific callable object.  If None or False, no progress updated will be
            displayed.  If True, a default progress bar will be created.
        force_download (bool, optional): download this file even if [destination_filename]
            exists.
        verbose (bool, optional): enable additional debug console output
        escape_spaces (bool, optional): replace ' ' with '%20'

    Returns:
        str: the filename to which [url] was downloaded, the same as [destination_filename]
        if [destination_filename] was not None
    """

    if progress_updater is not None and isinstance(progress_updater,bool):
        if not progress_updater:
            progress_updater = None
        else:
            progress_updater = DownloadProgressBar()

    url_no_sas = url.split('?')[0]

    if destination_filename is None:

        target_folder = make_temp_folder(subfolder='url_utils',append_guid=False)
        url_without_sas = url.split('?', 1)[0]

        # This does not guarantee uniqueness, hence "semi-best-effort"
        url_as_filename = re.sub(r'\W+', '', url_without_sas)

        n_folder_chars = len(target_folder)

        if (len(url_as_filename) + n_folder_chars) >= max_path_len:
            print('Warning: truncating filename target to {} characters'.format(max_path_len))
            max_fn_len = max_path_len - (n_folder_chars + 1)
            url_as_filename = url_as_filename[-1 * max_fn_len:]
        destination_filename = \
            os.path.join(target_folder,url_as_filename)

    # ...if the destination filename wasn't specified

    if escape_spaces:
        url = url.replace(' ','%20')

    if (not force_download) and (os.path.isfile(destination_filename)):
        if verbose:
            print('Bypassing download of already-downloaded file {}'.format(os.path.basename(url_no_sas)))
    else:
        if verbose:
            print('Downloading file {} to {}'.format(os.path.basename(url_no_sas),destination_filename),end='')
        target_dir = os.path.dirname(destination_filename)
        os.makedirs(target_dir,exist_ok=True)
        urllib.request.urlretrieve(url, destination_filename, progress_updater)
        assert(os.path.isfile(destination_filename))
        n_bytes = os.path.getsize(destination_filename)
        if verbose:
            print('...done, {} bytes.'.format(n_bytes))

    return destination_filename

# ...def download_url(...)


def download_relative_filename(url, output_base, verbose=False):
    """
    Download a URL to output_base, preserving relative path.  Path is relative to
    the site, so:

        https://abc.com/xyz/123.txt

    ...will get downloaded to:

        output_base/xyz/123.txt

    Args:
        url (str): the URL to download
        output_base (str): the base folder to which we should download this file
        verbose (bool, optional): enable additional debug console output

    Returns:
        str: the local destination filename
    """

    p = urlparse(url)
    # remove the leading '/'
    assert p.path.startswith('/'); relative_filename = p.path[1:]
    destination_filename = os.path.join(output_base,relative_filename)
    return download_url(url, destination_filename, verbose=verbose)

# ...def download_relative_filename(...)


def _do_parallelized_download(download_info,overwrite=False,verbose=False):
    """
    Internal function for download parallelization.
    """

    url = download_info['url']
    target_file = download_info['target_file']
    result = {'status':'unknown','url':url,'target_file':target_file}

    if ((os.path.isfile(target_file)) and (not overwrite)):
        if verbose:
            print('Skipping existing file {}'.format(target_file))
        result['status'] = 'skipped'
        return result
    try:
        download_url(url=url,
                     destination_filename=target_file,
                     verbose=verbose,
                     force_download=overwrite)
    except Exception as e:
        print('Warning: error downloading URL {}: {}'.format(
            url,str(e)))
        result['status'] = 'error: {}'.format(str(e))
        return result

    result['status'] = 'success'
    return result

# ...def _do_parallelized_download(...)


def parallel_download_urls(url_to_target_file,
                           verbose=False,
                           overwrite=False,
                           n_workers=20,
                           pool_type='thread'):
    """
    Downloads a list of URLs to local files.

    Catches exceptions and reports them in the returned "results" array.

    Args:
        url_to_target_file (dict): a dict mapping URLs to local filenames.
        verbose (bool, optional): enable additional debug console output
        overwrite (bool, optional): whether to overwrite existing local files
        n_workers (int, optional): number of concurrent workers, set to <=1 to disable
            parallelization
        pool_type (str, optional): worker type to use; should be 'thread' or 'process'

    Returns:
        list: list of dicts with keys:
            - 'url': the url this item refers to
            - 'status': 'skipped', 'success', or a string starting with 'error'
            - 'target_file': the local filename to which we downloaded (or tried to
              download) this URL
    """

    all_download_info = []

    if verbose:
        print('Preparing download list')
    for url in tqdm(url_to_target_file, disable=(not verbose)):
        download_info = {}
        download_info['url'] = url
        download_info['target_file'] = url_to_target_file[url]
        all_download_info.append(download_info)

    if verbose:
        print('Downloading {} images on {} workers'.format(
            len(all_download_info),n_workers))

    if n_workers <= 1:

        results = []

        for download_info in tqdm(all_download_info, disable=(not verbose)):
            result = _do_parallelized_download(download_info,overwrite=overwrite,verbose=verbose)
            results.append(result)

    else:

        pool = None

        try:
            if pool_type == 'thread':
                pool = ThreadPool(n_workers)
            else:
                assert pool_type == 'process', 'Unsupported pool type {}'.format(pool_type)
                pool = Pool(n_workers)

            if verbose:
                print('Starting a {} pool with {} workers'.format(pool_type,n_workers))

            results = list(tqdm(pool.imap(
                partial(_do_parallelized_download,overwrite=overwrite,verbose=verbose),
                all_download_info), total=len(all_download_info), disable=(not verbose)))

        finally:
            if pool:
                pool.close()
                pool.join()
                print("Pool closed and joined for parallel URL downloads")

    return results

# ...def parallel_download_urls(...)


@pytest.mark.skip(reason="This is not a test function")
def test_url(url,error_on_failure=True,timeout=None):
    """
    Tests the availability of [url], returning an http status code.

    Args:
        url (str): URL to test
        error_on_failure (bool, optional): whether to error (vs. just returning an
            error code) if accessing this URL fails
        timeout (int, optional): timeout in seconds to wait before considering this
            access attempt to be a failure; see requests.head() for precise documentation

    Returns:
        int: http status code (200 for success)
    """

    r = requests.head(url, stream=True, verify=True, timeout=timeout)

    if error_on_failure and r.status_code != 200:
        raise ValueError('Could not access {}: error {}'.format(url,r.status_code))
    return r.status_code


@pytest.mark.skip(reason="This is not a test function")
def test_urls(urls,error_on_failure=True,n_workers=1,pool_type='thread',timeout=None,verbose=False):
    """
    Verify that URLs are available (i.e., returns status 200).  By default,
    errors if any URL is unavailable.

    Args:
        urls (list): list of URLs to test
        error_on_failure (bool, optional): whether to error (vs. just returning an
            error code) if accessing this URL fails
        n_workers (int, optional): number of concurrent workers, set to <=1 to disable
            parallelization
        pool_type (str, optional): worker type to use; should be 'thread' or 'process'
        timeout (int, optional): timeout in seconds to wait before considering this
            access attempt to be a failure; see requests.head() for precise documentation
        verbose (bool, optional): enable additional debug output

    Returns:
        list: a list of http status codes, the same length and order as [urls]
    """

    if n_workers <= 1:

        status_codes = []

        for url in tqdm(urls,disable=(not verbose)):

            r = requests.get(url, timeout=timeout)

            if error_on_failure and r.status_code != 200:
                raise ValueError('Could not access {}: error {}'.format(url,r.status_code))
            status_codes.append(r.status_code)

    else:

        pool = None
        try:
            if pool_type == 'thread':
                pool = ThreadPool(n_workers)
            else:
                assert pool_type == 'process', 'Unsupported pool type {}'.format(pool_type)
                pool = Pool(n_workers)

            if verbose:
                print('Starting a {} pool with {} workers'.format(pool_type,n_workers))

            status_codes = list(tqdm(pool.imap(
                partial(test_url,error_on_failure=error_on_failure,timeout=timeout),
                urls), total=len(urls), disable=(not verbose)))
        finally:
            if pool:
                pool.close()
                pool.join()
                print('Pool closed and joined for URL tests')

    return status_codes

# ...def test_urls(...)


def get_url_size(url,verbose=False,timeout=None):
    """
    Get the size of the file pointed to by a URL, based on the Content-Length property.  If the
    URL is not available, or the Content-Length property is not available, or the content-Length
    property is not an integer, returns None.

    Args:
        url (str): the url to test
        verbose (bool, optional): enable additional debug output
        timeout (int, optional): timeout in seconds to wait before considering this
            access attempt to be a failure; see requests.head() for precise documentation

    Returns:
        int: the file size in bytes, or None if it can't be retrieved
    """

    try:
        r = urllib.request.Request(url,method='HEAD')
        f = urllib.request.urlopen(r, timeout=timeout)
        if f.status != 200:
            if verbose:
                print('Status {} retrieving file size for {}'.format(f.status,url))
            return None
        size_bytes_str = f.headers.get('Content-Length')
        if size_bytes_str is None:
            if verbose:
                print('No Content-Length header for {}'.format(url))
            return None
        size_bytes = int(size_bytes_str)
        return size_bytes
    except Exception as e:
        if verbose:
            print('Error retrieving file size for {}:\n{}'.format(url,str(e)))
        return None

# ...def get_url_size(...)


def get_url_sizes(urls,n_workers=1,pool_type='thread',timeout=None,verbose=False):
    """
    Retrieve file sizes for the URLs specified by [urls].  Returns None for any URLs
    that we can't access, or URLs for which the Content-Length property is not set.

    Args:
        urls (list): list of URLs for which we should retrieve sizes
        n_workers (int, optional): number of concurrent workers, set to <=1 to disable
            parallelization
        pool_type (str, optional): worker type to use; should be 'thread' or 'process'
        timeout (int, optional): timeout in seconds to wait before considering this
            access attempt to be a failure; see requests.head() for precise documentation
        verbose (bool, optional): print additional debug information

    Returns:
        dict: maps urls to file sizes, which will be None for URLs for which we were unable
        to retrieve a valid size.
    """

    url_to_size = {}

    if n_workers <= 1:

        for url in tqdm(urls, disable=(not verbose)):
            url_to_size[url] = get_url_size(url,verbose=verbose,timeout=timeout)

    else:

        pool = None
        try:
            if pool_type == 'thread':
                pool = ThreadPool(n_workers)
            else:
                assert pool_type == 'process', 'Unsupported pool type {}'.format(pool_type)
                pool = Pool(n_workers)

            if verbose:
                print('Starting a {} pool with {} workers'.format(pool_type,n_workers))

            file_sizes = list(tqdm(pool.imap(
                partial(get_url_size,verbose=verbose,timeout=timeout),
                urls), total=len(urls), disable=(not verbose)))

            for i_url,url in enumerate(urls):
                url_to_size[url] = file_sizes[i_url]
        finally:
            if pool:
                pool.close()
                pool.join()
                print('Pool closed and joined for URL size checks')

    return url_to_size


#%% Tests

# Constants for tests

SMALL_FILE_URL = "https://www.google.com/images/branding/googlelogo/1x/googlelogo_color_272x92dp.png"
REDIRECT_SRC_URL = "http://google.com"
REDIRECT_DEST_URL = "https://www.google.com/"
NON_EXISTENT_URL = "https://example.com/non_existent_page_404.html"
DEFINITELY_NON_EXISTENT_DOMAIN_URL = "https://thisshouldnotexist1234567890.com/file.txt"
RELATIVE_DOWNLOAD_URL = "https://raw.githubusercontent.com/agentmorris/MegaDetector/main/README.md"
RELATIVE_DOWNLOAD_CONTAIN_TOKEN = 'agentmorris'
RELATIVE_DOWNLOAD_NOT_CONTAIN_TOKEN = 'github'


class TestUrlUtils:
    """
    Tests for url_utils.py
    """

    def set_up(self):
        """
        Create a temporary directory for testing.
        """

        self.test_dir = make_test_folder(subfolder='url_utils_tests')
        self.download_target_dir = os.path.join(self.test_dir, 'downloads')
        os.makedirs(self.download_target_dir, exist_ok=True)


    def tear_down(self):
        """
        Remove the temporary directory after tests and restore module temp_dir.
        """

        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)


    def test_download_url_to_specified_file(self):
        """
        Test download_url with a specified destination filename.
        """

        dest_filename = os.path.join(self.download_target_dir, "downloaded_google_logo.png")
        returned_filename = download_url(SMALL_FILE_URL,
                                         destination_filename=dest_filename,
                                         verbose=False)
        assert returned_filename == dest_filename
        assert os.path.exists(dest_filename)
        assert os.path.getsize(dest_filename) > 1000


    def test_download_url_to_temp_file(self):
        """
        Test download_url when destination_filename is None.
        """

        returned_filename = download_url(SMALL_FILE_URL,
                                         destination_filename=None,
                                         verbose=False)
        assert os.path.exists(returned_filename)
        assert os.path.getsize(returned_filename) > 1000


    def test_download_url_non_existent(self):
        """
        Test download_url with a non-existent URL.
        """

        dest_filename = os.path.join(self.download_target_dir, "non_existent.html")
        try:
            download_url(NON_EXISTENT_URL, destination_filename=dest_filename, verbose=False)
            raise AssertionError("urllib.error.HTTPError not raised for 404")
        except urllib.error.HTTPError:
            pass

        try:
            download_url(DEFINITELY_NON_EXISTENT_DOMAIN_URL,
                         destination_filename=dest_filename,
                         verbose=False)
            raise AssertionError(
                "urllib.error.URLError or requests.exceptions.ConnectionError not raised for DNS failure")
        except urllib.error.URLError:
            pass
        except requests.exceptions.ConnectionError:
            pass


    def test_download_url_force_download(self):
        """
        Test the force_download parameter of download_url.
        """

        dest_filename = os.path.join(self.download_target_dir, "force_test.png")

        download_url(SMALL_FILE_URL, destination_filename=dest_filename, verbose=False)
        assert os.path.exists(dest_filename)
        initial_mtime = os.path.getmtime(dest_filename)

        download_url(SMALL_FILE_URL, destination_filename=dest_filename, verbose=True)
        assert os.path.getmtime(dest_filename) == initial_mtime

        download_url(SMALL_FILE_URL,
                     destination_filename=dest_filename,
                     force_download=True,
                     verbose=False)
        assert os.path.exists(dest_filename)


    def test_download_url_escape_spaces(self):
        """
        Test download_url with spaces in the URL.
        """

        dest_filename = os.path.join(self.download_target_dir, "escape_test.png")
        download_url(SMALL_FILE_URL,
                     destination_filename=dest_filename,
                     escape_spaces=True,
                     verbose=False)
        assert os.path.exists(dest_filename)


    def test_download_relative_filename(self):
        """
        Test download_relative_filename.
        """

        output_base = os.path.join(self.download_target_dir, "relative_dl")
        returned_filename = download_relative_filename(RELATIVE_DOWNLOAD_URL, output_base, verbose=False)
        assert RELATIVE_DOWNLOAD_CONTAIN_TOKEN in returned_filename
        assert RELATIVE_DOWNLOAD_NOT_CONTAIN_TOKEN not in returned_filename
        assert os.path.exists(returned_filename)
        assert os.path.getsize(returned_filename) > 100


    def test_parallel_download_urls(self):
        """
        Test parallel_download_urls (with n_workers=1 for simplicity).
        """

        url1_target = os.path.join(self.download_target_dir, "parallel_dl_1.png")
        url2_target = os.path.join(self.download_target_dir, "parallel_dl_2_nonexistent.html")

        url_to_target_file = {
            SMALL_FILE_URL: url1_target,
            NON_EXISTENT_URL: url2_target
        }

        results = parallel_download_urls(url_to_target_file, n_workers=1, verbose=False)

        assert len(results) == 2

        status_map = {res['url']: res for res in results}

        assert status_map[SMALL_FILE_URL]['status'] == 'success'
        assert status_map[SMALL_FILE_URL]['target_file'] == url1_target
        assert os.path.exists(url1_target)

        assert status_map[NON_EXISTENT_URL]['status'].startswith('error: HTTP Error 404')
        assert status_map[NON_EXISTENT_URL]['target_file'] == url2_target
        assert not os.path.exists(url2_target)

        if not os.path.exists(url1_target):
             download_url(SMALL_FILE_URL, url1_target, verbose=False)
        results_skip = parallel_download_urls({SMALL_FILE_URL: url1_target},
                                              n_workers=1,
                                              overwrite=False,
                                              verbose=True)
        assert results_skip[0]['status'] == 'skipped'

        results_overwrite = parallel_download_urls({SMALL_FILE_URL: url1_target},
                                                   n_workers=1,
                                                   overwrite=True,
                                                   verbose=False)
        assert results_overwrite[0]['status'] == 'success'


    def test_test_url_and_test_urls(self):
        """
        Test test_url and test_urls functions.
        """

        assert test_url(SMALL_FILE_URL, error_on_failure=False, timeout=10) == 200
        assert test_url(REDIRECT_SRC_URL, error_on_failure=False, timeout=10) in (200,301)

        status_non_existent = test_url(NON_EXISTENT_URL, error_on_failure=False, timeout=5)
        assert status_non_existent == 404

        try:
            test_url(NON_EXISTENT_URL, error_on_failure=True, timeout=5)
            raise AssertionError("ValueError not raised for NON_EXISTENT_URL")
        except ValueError:
            pass

        try:
            test_url(DEFINITELY_NON_EXISTENT_DOMAIN_URL,
                     error_on_failure=True,
                     timeout=5)
            raise AssertionError("requests.exceptions.ConnectionError or urllib.error.URLError not raised")
        except requests.exceptions.ConnectionError:
            pass
        except urllib.error.URLError:
            pass


        urls_to_test = [SMALL_FILE_URL, NON_EXISTENT_URL]
        status_codes = test_urls(urls_to_test, error_on_failure=False, n_workers=1, timeout=10)
        assert len(status_codes) == 2
        assert status_codes[0] == 200
        assert status_codes[1] == 404

        try:
            test_urls(urls_to_test, error_on_failure=True, n_workers=1, timeout=5)
            raise AssertionError("ValueError not raised for urls_to_test")
        except ValueError:
            pass

        good_urls = [SMALL_FILE_URL, REDIRECT_SRC_URL]
        good_status_codes = test_urls(good_urls, error_on_failure=True, n_workers=1, timeout=10)
        assert good_status_codes == [200, 200]


    def test_get_url_size_and_sizes(self):
        """
        Test get_url_size and get_url_sizes functions.
        """

        size = get_url_size(SMALL_FILE_URL, timeout=10)
        assert size is not None
        assert size > 1000

        size_dynamic = get_url_size(REDIRECT_DEST_URL, timeout=10, verbose=True)
        if size_dynamic is not None:
            assert isinstance(size_dynamic, int)

        size_non_existent = get_url_size(NON_EXISTENT_URL, timeout=5)
        assert size_non_existent is None

        size_bad_domain = get_url_size(DEFINITELY_NON_EXISTENT_DOMAIN_URL, timeout=5)
        assert size_bad_domain is None

        urls_for_size = [SMALL_FILE_URL, NON_EXISTENT_URL, REDIRECT_DEST_URL]
        sizes_map = get_url_sizes(urls_for_size, n_workers=1, timeout=10)

        assert SMALL_FILE_URL in sizes_map
        assert sizes_map[SMALL_FILE_URL] == size

        assert NON_EXISTENT_URL in sizes_map
        assert sizes_map[NON_EXISTENT_URL] is None

        assert REDIRECT_DEST_URL in sizes_map
        assert sizes_map[REDIRECT_DEST_URL] == size_dynamic


def _test_url_utils():
    """
    Runs all tests in the TestUrlUtils class.  I generally disable this during testing
    because it creates irritating nondeterminism, and this is neither a core module nor
    a module that changes often.
    """

    test_instance = TestUrlUtils()
    test_instance.set_up()
    try:
        test_instance.test_download_url_to_specified_file()
        test_instance.test_download_url_to_temp_file()
        test_instance.test_download_url_non_existent()
        test_instance.test_download_url_force_download()
        test_instance.test_download_url_escape_spaces()
        test_instance.test_download_relative_filename()
        test_instance.test_parallel_download_urls()
        test_instance.test_test_url_and_test_urls()
        test_instance.test_get_url_size_and_sizes()
    finally:
        test_instance.tear_down()

# from IPython import embed; embed()
# test_url_utils()



"""

render_detection_confusion_matrix.py

Given a CCT-formatted ground truth file and a MegaDetector-formatted results file,
render an HTML confusion matrix.  Typically used for multi-class detectors.  Currently
assumes a single class per image.

"""

#%% Imports and constants

import os
import json

import matplotlib.pyplot as plt
import numpy as np

from tqdm import tqdm
from collections import defaultdict
from functools import partial

from megadetector.utils.path_utils import find_images
from megadetector.utils.path_utils import flatten_path
from megadetector.utils.write_html_image_list import write_html_image_list
from megadetector.visualization import visualization_utils as vis_utils
from megadetector.visualization import plot_utils

from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool


#%% Support functions

def _image_to_output_file(im,preview_images_folder):
    """
    Produces a clean filename from im (if [im] is a str) or im['file'] (if [im] is a dict).
    """

    if isinstance(im,str):
        filename_relative = im
    else:
        filename_relative = im['file']

    fn_clean = flatten_path(filename_relative).replace(' ','_')
    return os.path.join(preview_images_folder,fn_clean)


def _render_image(im,render_image_constants):
    """
    Internal function for rendering a single image to the confusion matrix preview folder.
    """

    filename_to_ground_truth_im = render_image_constants['filename_to_ground_truth_im']
    image_folder = render_image_constants['image_folder']
    preview_images_folder = render_image_constants['preview_images_folder']
    force_render_images = render_image_constants['force_render_images']
    results_category_id_to_name = render_image_constants['results_category_id_to_name']
    rendering_confidence_thresholds = render_image_constants['rendering_confidence_thresholds']
    target_image_size = render_image_constants['target_image_size']

    assert im['file'] in filename_to_ground_truth_im

    output_file = _image_to_output_file(im,preview_images_folder)
    if os.path.isfile(output_file) and not force_render_images:
        return output_file

    input_file = os.path.join(image_folder,im['file'])
    assert os.path.isfile(input_file)

    detections_to_render = []

    for det in im['detections']:
        category_name = results_category_id_to_name[det['category']]
        detection_threshold = rendering_confidence_thresholds['default']
        if category_name in rendering_confidence_thresholds:
            detection_threshold = rendering_confidence_thresholds[category_name]
        if det['conf'] > detection_threshold:
            detections_to_render.append(det)

    vis_utils.draw_bounding_boxes_on_file(input_file, output_file, detections_to_render,
                                          detector_label_map=results_category_id_to_name,
                                          label_font_size=20,target_size=target_image_size)

    return output_file


#%% Main function

def render_detection_confusion_matrix(ground_truth_file,
                                      results_file,
                                      image_folder,
                                      preview_folder,
                                      force_render_images=False,
                                      confidence_thresholds=None,
                                      rendering_confidence_thresholds=None,
                                      target_image_size=(1280,-1),
                                      parallelize_rendering=True,
                                      parallelize_rendering_n_cores=None,
                                      parallelize_rendering_with_threads=False,
                                      job_name='unknown',
                                      model_file=None,
                                      empty_category_name='empty',
                                      html_image_list_options=None):
    """
    Given a CCT-formatted ground truth file and a MegaDetector-formatted results file,
    render an HTML confusion matrix in [preview_folder.  Typically used for multi-class detectors.
    Currently assumes a single class per image.

    confidence_thresholds and rendering_confidence_thresholds are dictionaries mapping
    class names to thresholds.  "default" is a special token that will be used for all
    classes not otherwise assigned thresholds.

    Args:
        ground_truth_file (str): the CCT-formatted .json file with ground truth information
        results_file (str): the MegaDetector results .json file
        image_folder (str): the folder where images live; filenames in [ground_truth_file] and
            [results_file] should be relative to this folder.
        preview_folder (str): the output folder, i.e. the folder in which we'll create our nifty
            HTML stuff.
        force_render_images (bool, optional): if False, skips images that already exist
        confidence_thresholds (dict, optional): a dictionary mapping class names to thresholds;
            all classes not explicitly named here will use the threshold for the "default" category.
        rendering_confidence_thresholds (dict, optional): a dictionary mapping class names to thresholds;
            all classes not explicitly named here will use the threshold for the "default" category.
        target_image_size (tuple, optional): output image size, as a pair of ints (width,height).  If one
            value is -1 and the other is not, aspect ratio is preserved.  If both are -1, the original image
            sizes are preserved.
        parallelize_rendering (bool, optional): enable (default) or disable parallelization when rendering
        parallelize_rendering_n_cores (int, optional): number of threads or processes to use for rendering, only
            used if parallelize_rendering is True
        parallelize_rendering_with_threads (bool, optional): whether to use threads (True) or processes (False)
            when rendering, only used if parallelize_rendering is True
        job_name (str, optional): job name to include in big letters in the output file
        model_file (str, optional): model filename to include in HTML output
        empty_category_name (str, optional): special category name that we should treat as empty, typically
            "empty"
        html_image_list_options (dict, optional): options listed passed along to write_html_image_list;
            see write_html_image_list for documentation.

    Returns:
        dict: confusion matrix information, containing at least the key "html_file"
    """

    ##%% Argument and path handling

    preview_images_folder = os.path.join(preview_folder,'images')
    os.makedirs(preview_images_folder,exist_ok=True)

    if confidence_thresholds is None:
        confidence_thresholds = {'default':0.5}
    if rendering_confidence_thresholds is None:
        rendering_confidence_thresholds = {'default':0.4}


    ##%% Load ground truth

    with open(ground_truth_file,'r') as f:
        ground_truth_data_cct = json.load(f)

    filename_to_ground_truth_im = {}
    for im in ground_truth_data_cct['images']:
        assert im['file_name'] not in filename_to_ground_truth_im
        filename_to_ground_truth_im[im['file_name']] = im


    ##%% Confirm that the ground truth images are present in the image folder

    ground_truth_images = find_images(image_folder,return_relative_paths=True,recursive=True)
    assert len(ground_truth_images) == len(ground_truth_data_cct['images'])
    del ground_truth_images


    ##%% Map images to categories

    # gt_image_id_to_image = {im['id']:im for im in ground_truth_data_cct['images']}
    gt_image_id_to_annotations = defaultdict(list)

    ground_truth_category_id_to_name = {}
    for c in ground_truth_data_cct['categories']:
        ground_truth_category_id_to_name[c['id']] = c['name']

    # Add the empty category if necessary
    if empty_category_name not in ground_truth_category_id_to_name.values():
        empty_category_id = max(ground_truth_category_id_to_name.keys())+1
        ground_truth_category_id_to_name[empty_category_id] = empty_category_name

    ground_truth_category_names = sorted(list(ground_truth_category_id_to_name.values()))

    for ann in ground_truth_data_cct['annotations']:
        gt_image_id_to_annotations[ann['image_id']].append(ann)

    gt_filename_to_category_names = defaultdict(set)

    for im in ground_truth_data_cct['images']:
        annotations_this_image = gt_image_id_to_annotations[im['id']]
        for ann in annotations_this_image:
            category_name = ground_truth_category_id_to_name[ann['category_id']]
            gt_filename_to_category_names[im['file_name']].add(category_name)

    for filename in gt_filename_to_category_names:

        category_names_this_file = gt_filename_to_category_names[filename]

        # The empty category should be exclusive
        if empty_category_name in category_names_this_file:
            assert len(category_names_this_file) == 1, \
              'Empty category assigned along with another category for {}'.format(filename)
        assert len(category_names_this_file) > 0, \
            'No ground truth category assigned to {}'.format(filename)


    ##%% Load results

    with open(results_file,'r') as f:
        md_formatted_results = json.load(f)

    results_category_id_to_name = md_formatted_results['detection_categories']


    ##%% Render images with detections

    render_image_constants = {}
    render_image_constants['filename_to_ground_truth_im'] = filename_to_ground_truth_im
    render_image_constants['image_folder'] = image_folder
    render_image_constants['preview_images_folder'] = preview_images_folder
    render_image_constants['force_render_images'] = force_render_images
    render_image_constants['results_category_id_to_name'] = results_category_id_to_name
    render_image_constants['rendering_confidence_thresholds'] = rendering_confidence_thresholds
    render_image_constants['target_image_size'] = target_image_size

    if parallelize_rendering:

        pool = None
        try:
            if parallelize_rendering_n_cores is None:
                if parallelize_rendering_with_threads:
                    pool = ThreadPool()
                else:
                    pool = Pool()
            else:
                if parallelize_rendering_with_threads:
                    pool = ThreadPool(parallelize_rendering_n_cores)
                    worker_string = 'threads'
                else:
                    pool = Pool(parallelize_rendering_n_cores)
                    worker_string = 'processes'
                print('Rendering images with {} {}'.format(parallelize_rendering_n_cores,
                                                        worker_string))

            _ = list(tqdm(pool.imap(partial(_render_image,render_image_constants=render_image_constants),
                                    md_formatted_results['images']),
                                    total=len(md_formatted_results['images'])))
        finally:
            pool.close()
            pool.join()
            print("Pool closed and joined for confusion matrix rendering")

    else:

        # im = md_formatted_results['images'][0]
        for im in tqdm(md_formatted_results['images']):
            _render_image(im,render_image_constants)


    ##%% Map images to predicted categories, and vice-versa

    filename_to_predicted_categories = defaultdict(set)
    predicted_category_name_to_filenames = defaultdict(set)

    # im = md_formatted_results['images'][0]
    for im in tqdm(md_formatted_results['images']):

        assert im['file'] in filename_to_ground_truth_im

        # det = im['detections'][0]
        for det in im['detections']:
            category_name = results_category_id_to_name[det['category']]
            detection_threshold = confidence_thresholds['default']
            if category_name in confidence_thresholds:
                detection_threshold = confidence_thresholds[category_name]
            if det['conf'] > detection_threshold:
                filename_to_predicted_categories[im['file']].add(category_name)
                predicted_category_name_to_filenames[category_name].add(im['file'])

        # ...for each detection

    # ...for each image


    ##%% Create TP/TN/FP/FN lists

    category_name_to_image_lists = {}

    sub_page_tokens = ['fn','tn','fp','tp']

    for category_name in ground_truth_category_names:

        category_name_to_image_lists[category_name] = {}
        for sub_page_token in sub_page_tokens:
            category_name_to_image_lists[category_name][sub_page_token] = []

    # filename = next(iter(gt_filename_to_category_names))
    for filename in gt_filename_to_category_names.keys():

        ground_truth_categories_this_image = gt_filename_to_category_names[filename]
        predicted_categories_this_image = filename_to_predicted_categories[filename]

        for category_name in ground_truth_category_names:

            assignment = None

            if category_name == empty_category_name:
                # If this is an empty image
                if category_name in ground_truth_categories_this_image:
                    assert len(ground_truth_categories_this_image) == 1
                    if len(predicted_categories_this_image) == 0:
                        assignment = 'tp'
                    else:
                        assignment = 'fn'
                # If this not an empty image
                else:
                    if len(predicted_categories_this_image) == 0:
                        assignment = 'fp'
                    else:
                        assignment = 'tn'

            else:
                if category_name in ground_truth_categories_this_image:
                    if category_name in predicted_categories_this_image:
                        assignment = 'tp'
                    else:
                        assignment = 'fn'
                else:
                    if category_name in predicted_categories_this_image:
                        assignment = 'fp'
                    else:
                        assignment = 'tn'

            category_name_to_image_lists[category_name][assignment].append(filename)

    # ...for each filename


    ##%% Create confusion matrix

    gt_category_name_to_category_index = {}

    for i_category,category_name in enumerate(ground_truth_category_names):
        gt_category_name_to_category_index[category_name] = i_category

    n_categories = len(gt_category_name_to_category_index)

    # indexed as [true,predicted]
    confusion_matrix = np.zeros(shape=(n_categories,n_categories),dtype=int)

    filename_to_results_im = {im['file']:im for im in md_formatted_results['images']}

    true_predicted_to_file_list = defaultdict(list)

    # filename = next(iter(gt_filename_to_category_names.keys()))
    for filename in gt_filename_to_category_names.keys():

        ground_truth_categories_this_image = gt_filename_to_category_names[filename]
        assert len(ground_truth_categories_this_image) == 1
        ground_truth_category_name = next(iter(ground_truth_categories_this_image))

        results_im = filename_to_results_im[filename]

        # If there were no detections at all, call this image empty
        if len(results_im['detections']) == 0:
            predicted_category_name = empty_category_name
        # Otherwise look for above-threshold detections
        else:
            results_category_name_to_confidence = defaultdict(int)
            for det in results_im['detections']:
                category_name = results_category_id_to_name[det['category']]
                detection_threshold = confidence_thresholds['default']
                if category_name in confidence_thresholds:
                    detection_threshold = confidence_thresholds[category_name]
                if det['conf'] > detection_threshold:
                    results_category_name_to_confidence[category_name] = max(
                        results_category_name_to_confidence[category_name],det['conf'])
                # If there were no detections above threshold
                if len(results_category_name_to_confidence) == 0:
                    predicted_category_name = empty_category_name
                else:
                    predicted_category_name = max(results_category_name_to_confidence,
                        key=results_category_name_to_confidence.get)

        ground_truth_category_index = gt_category_name_to_category_index[ground_truth_category_name]
        predicted_category_index = gt_category_name_to_category_index[predicted_category_name]

        true_predicted_token = ground_truth_category_name + '_' + predicted_category_name
        true_predicted_to_file_list[true_predicted_token].append(filename)

        confusion_matrix[ground_truth_category_index,predicted_category_index] += 1

    # ...for each file

    plt.ioff()

    fig_h = 3 + 0.3 * n_categories
    fig_w = fig_h
    fig = plt.figure(figsize=(fig_w, fig_h),tight_layout=True)

    plot_utils.plot_confusion_matrix(
        matrix=confusion_matrix,
        classes=ground_truth_category_names,
        normalize=False,
        title='Confusion matrix',
        cmap=plt.cm.Blues,
        vmax=1.0,
        use_colorbar=False,
        y_label=True,
        fig=fig)

    cm_figure_fn_relative = 'confusion_matrix.png'
    cm_figure_fn_abs = os.path.join(preview_folder, cm_figure_fn_relative)
    # fig.show()
    fig.savefig(cm_figure_fn_abs,dpi=100)
    plt.close(fig)

    # open_file(cm_figure_fn_abs)


    ##%% Create HTML confusion matrix

    html_confusion_matrix = '<table class="result-table">\n'
    html_confusion_matrix += '<tr>\n'
    html_confusion_matrix += '<td>{}</td>\n'.format('True category')
    for category_name in ground_truth_category_names:
        html_confusion_matrix += '<td>{}</td>\n'.format('&nbsp;')
    html_confusion_matrix += '</tr>\n'

    for true_category in ground_truth_category_names:

        html_confusion_matrix += '<tr>\n'
        html_confusion_matrix += '<td>{}</td>\n'.format(true_category)

        for predicted_category in ground_truth_category_names:

            true_predicted_token = true_category + '_' + predicted_category
            image_list = true_predicted_to_file_list[true_predicted_token]
            if len(image_list) == 0:
                td_content = '0'
            else:
                if html_image_list_options is None:
                    html_image_list_options = {}
                title_string = 'true: {}, predicted {}'.format(
                    true_category,predicted_category)
                html_image_list_options['headerHtml'] = '<h1>{}</h1>'.format(title_string)

                html_image_info_list = []

                for image_filename_relative in image_list:
                    html_image_info = {}
                    detections = filename_to_results_im[image_filename_relative]['detections']
                    if len(detections) == 0:
                        max_conf = 0
                    else:
                        max_conf = max([d['conf'] for d in detections])

                    title = '<b>Image</b>: {}, <b>Max conf</b>: {:0.3f}'.format(
                        image_filename_relative, max_conf)
                    image_link = 'images/' + os.path.basename(
                        _image_to_output_file(image_filename_relative,preview_images_folder))
                    html_image_info = {
                        'filename': image_link,
                        'title': title,
                        'textStyle':\
                         'font-family:verdana,arial,calibri;font-size:80%;' + \
                             'text-align:left;margin-top:20;margin-bottom:5'
                    }

                    html_image_info_list.append(html_image_info)

                target_html_file_relative = true_predicted_token + '.html'
                target_html_file_abs = os.path.join(preview_folder,target_html_file_relative)
                write_html_image_list(
                    filename=target_html_file_abs,
                    images=html_image_info_list,
                    options=html_image_list_options)

                td_content = '<a href="{}">{}</a>'.format(target_html_file_relative,
                                                          len(image_list))

            html_confusion_matrix += '<td>{}</td>\n'.format(td_content)

        # ...for each predicted category

        html_confusion_matrix += '</tr>\n'

    # ...for each true category

    html_confusion_matrix += '<tr>\n'
    html_confusion_matrix += '<td>&nbsp;</td>\n'

    for category_name in ground_truth_category_names:
        html_confusion_matrix += '<td class="rotate"><p style="margin-left:20px;">{}</p></td>\n'.format(
            category_name)
    html_confusion_matrix += '</tr>\n'

    html_confusion_matrix += '</table>'


    ##%% Create HTML sub-pages and HTML table

    html_table = '<table class="result-table">\n'

    html_table += '<tr>\n'
    html_table += '<td>{}</td>\n'.format('True category')
    for sub_page_token in sub_page_tokens:
        html_table += '<td>{}</td>'.format(sub_page_token)
    html_table += '</tr>\n'

    filename_to_results_im = {im['file']:im for im in md_formatted_results['images']}

    sub_page_token_to_page_name = {
        'fp':'false positives',
        'tp':'true positives',
        'fn':'false negatives',
        'tn':'true negatives'
    }

    # category_name = ground_truth_category_names[0]
    for category_name in ground_truth_category_names:

        html_table += '<tr>\n'

        html_table += '<td>{}</td>\n'.format(category_name)

        # sub_page_token = sub_page_tokens[0]
        for sub_page_token in sub_page_tokens:

            html_table += '<td>\n'

            image_list = category_name_to_image_lists[category_name][sub_page_token]

            if len(image_list) == 0:

                html_table += '0\n'

            else:

                html_image_list_options = {}
                title_string = '{}: {}'.format(category_name,sub_page_token_to_page_name[sub_page_token])
                html_image_list_options['headerHtml'] = '<h1>{}</h1>'.format(title_string)

                target_html_file_relative = '{}_{}.html'.format(category_name,sub_page_token)
                target_html_file_abs = os.path.join(preview_folder,target_html_file_relative)

                html_image_info_list = []

                # image_filename_relative = image_list[0]
                for image_filename_relative in image_list:

                    source_file = os.path.join(image_folder,image_filename_relative)
                    assert os.path.isfile(source_file)

                    html_image_info = {}
                    detections = filename_to_results_im[image_filename_relative]['detections']
                    if len(detections) == 0:
                        max_conf = 0
                    else:
                        max_conf = max([d['conf'] for d in detections])

                    title = '<b>Image</b>: {}, <b>Max conf</b>: {:0.3f}'.format(
                        image_filename_relative, max_conf)
                    image_link = 'images/' + os.path.basename(
                        _image_to_output_file(image_filename_relative,preview_images_folder))
                    html_image_info = {
                        'filename': image_link,
                        'title': title,
                        'linkTarget': source_file,
                        'textStyle':\
                         'font-family:verdana,arial,calibri;font-size:80%;' + \
                             'text-align:left;margin-top:20;margin-bottom:5'
                    }

                    html_image_info_list.append(html_image_info)

                # ...for each image

                write_html_image_list(
                    filename=target_html_file_abs,
                    images=html_image_info_list,
                    options=html_image_list_options)

                html_table += '<a href="{}">{}</a>\n'.format(target_html_file_relative,len(image_list))

            html_table += '</td>\n'

        # ...for each sub-page

        html_table += '</tr>\n'

    # ...for each category

    html_table += '</table>'

    html = '<html>\n'

    style_header = """<head>
        <style type="text/css">
        a { text-decoration: none; }
        body { font-family: segoe ui, calibri, "trebuchet ms", verdana, arial, sans-serif; }
        div.contentdiv { margin-left: 20px; }
        table.result-table { border:1px solid black; border-collapse: collapse; margin-left:50px;}
        td,th { padding:10px; }
        .rotate {
          padding:0px;
          writing-mode:vertical-lr;
          -webkit-transform: rotate(-180deg);
          -moz-transform: rotate(-180deg);
          -ms-transform: rotate(-180deg);
          -o-transform: rotate(-180deg);
          transform: rotate(-180deg);
        }
        </style>
        </head>"""

    html += style_header + '\n'

    html += '<body>\n'

    html += '<h1>Results summary for {}</h1>\n'.format(job_name)

    if model_file is not None and len(model_file) > 0:
        html += '<p><b>Model file</b>: {}</p>'.format(os.path.basename(model_file))

    html += '<p><b>Confidence thresholds</b></p>'

    for c in confidence_thresholds.keys():
        html += '<p style="margin-left:15px;">{}: {}</p>'.format(c,confidence_thresholds[c])

    html += '<h2>Confusion matrix</h2>\n'

    html += '<p>...assuming a single category per image.</p>\n'

    html += '<img src="{}"/>\n'.format(cm_figure_fn_relative)

    html += '<h2>Confusion matrix (with links)</h2>\n'

    html += '<p>...assuming a single category per image.</p>\n'

    html += html_confusion_matrix

    html += '<h2>Per-class statistics</h2>\n'

    html += html_table

    html += '</body>\n'
    html += '<html>\n'

    target_html_file = os.path.join(preview_folder,'index.html')

    with open(target_html_file,'w') as f:
        f.write(html)


    ##%% Prepare return data

    confusion_matrix_info = {}
    confusion_matrix_info['html_file'] = target_html_file

    return confusion_matrix_info

# ...render_detection_confusion_matrix(...)



"""

add_width_and_height_to_db.py

Grabs width and height from actual image files for a .json database that is missing w/h.

"""

#%% Imports and constants

import os
import sys
import json
import argparse

from tqdm import tqdm
from PIL import Image

from megadetector.utils import ct_utils


#%% Main resizing function

def add_width_and_height_to_db(input_file,output_file,image_base_folder):
    """
    Add width and height to images in the COCO db [input_file]
    that don't have non-None w/h values.  Does not verify correctness
    for images that already have non-None w/h values.  Ignores files that
    fail to open.

    Args:
        input_file (str): the COCO .json file to process
        output_file (str): the COCO .json file to write
        image_base_folder (str): image filenames in [input_file] should be relative
            to this folder

    Returns:
        list: the list of image dicts that were modified
    """

    with open(input_file,'r') as f:
        d = json.load(f)

    to_return = []

    for im in tqdm(d['images']):

        if ('height' not in im) or ('width' not in im) or \
           (im['height'] is None) or (im['width'] is None) or \
           (im['height'] <= 0) or (im['width'] <= 0):

            fn_relative = im['file_name']
            fn_abs = os.path.join(image_base_folder,fn_relative)

            if not os.path.isfile(fn_abs):
                print('Could not find image file {}'.format(fn_abs))
                continue

            try:
                im_w, im_h = Image.open(fn_abs).size
            except Exception as e:
                print('Error opening file {}: {}'.format(fn_abs,str(e)))
                continue

            assert isinstance(im_w,int) and isinstance(im_h,int) and \
                im_w > 0 and im_h > 0, \
                'Illegal size retrieved for {}'.format(fn_abs)

            im['height'] = im_h
            im['width'] = im_w
            to_return.append(im)

        # ...if we need to add width and/or height to this image

    # ...for each image

    ct_utils.write_json(output_file, d)

    print('Added size information to {} of {} images'.format(
        len(to_return), len(d['images'])))

    return to_return

# ...def add_width_and_height_to_db(...)


#%% Command-line driver

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=str,
                        help='Input COCO-formatted .json file')
    parser.add_argument('output_file', type=str,
                        help='Output COCO-formatted .json file')
    parser.add_argument('image_base_folder', type=str,
                        help='Base directory for images')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    add_width_and_height_to_db(args.input_file,
                               args.output_file,
                               args.image_base_folder)



"""

lila_common.py

Common constants and functions related to LILA data management/retrieval.

"""

#%% Imports and constants

import os
import json
import zipfile
import pandas as pd

from urllib.parse import urlparse

from megadetector.utils.url_utils import download_url
from megadetector.utils.path_utils import unzip_file
from megadetector.utils.ct_utils import is_empty

# LILA camera trap primary metadata file
lila_metadata_url = 'http://lila.science/wp-content/uploads/2023/06/lila_camera_trap_datasets.csv'
lila_taxonomy_mapping_url = 'https://lila.science/public/lila-taxonomy-mapping_release.csv'
lila_all_images_url = 'https://lila.science/public/lila_image_urls_and_labels.csv.zip'

wildlife_insights_page_size = 30000
wildlife_insights_taxonomy_url = 'https://api.wildlifeinsights.org/api/v1/taxonomy/taxonomies-all?fields=class,order,family,genus,species,authority,taxonomyType,uniqueIdentifier,commonNameEnglish&page[size]={}'.format(
    wildlife_insights_page_size)
wildlife_insights_taxonomy_local_json_filename = 'wi_taxonomy.json'
wildlife_insights_taxonomy_local_csv_filename = \
    wildlife_insights_taxonomy_local_json_filename.replace('.json','.csv')

# Filenames are consistent across clouds relative to these URLs
lila_base_urls = {
    'azure':'https://lilawildlife.blob.core.windows.net/lila-wildlife/',
    'gcp':'https://storage.googleapis.com/public-datasets-lila/',
    'aws':'http://us-west-2.opendata.source.coop.s3.amazonaws.com/agentmorris/lila-wildlife/'
}

lila_cloud_urls = {
    'azure':'https://lilawildlife.blob.core.windows.net/lila-wildlife/',
    'gcp':'gs://public-datasets-lila/',
    'aws':'s3://us-west-2.opendata.source.coop/agentmorris/lila-wildlife/'
}

for url in lila_base_urls.values():
    assert url.endswith('/')


#%% Common functions

def read_wildlife_insights_taxonomy_mapping(metadata_dir, force_download=False):
    """
    Reads the WI taxonomy mapping file, downloading the .json data (and writing to .csv) if necessary.

    Args:
        metadata_dir (str): folder to use for temporary LILA metadata files
        force_download (bool, optional): download the taxonomy mapping file
            even if the local file exists.

    Returns:
        pd.dataframe: A DataFrame with taxonomy information
    """

    wi_taxonomy_csv_path = os.path.join(metadata_dir,wildlife_insights_taxonomy_local_csv_filename)

    if os.path.exists(wi_taxonomy_csv_path):
        df = pd.read_csv(wi_taxonomy_csv_path)
    else:
        wi_taxonomy_json_path = os.path.join(metadata_dir,wildlife_insights_taxonomy_local_json_filename)
        download_url(wildlife_insights_taxonomy_url, wi_taxonomy_json_path,
                     force_download=force_download)
        with open(wi_taxonomy_json_path,'r') as f:
            d = json.load(f)

        # We haven't implemented paging, make sure that's not an issue
        assert d['meta']['totalItems'] < wildlife_insights_page_size

        # d['data'] is a list of items that look like:
        """
         {'id': 2000003,
         'class': 'Mammalia',
         'order': 'Rodentia',
         'family': 'Abrocomidae',
         'genus': 'Abrocoma',
         'species': 'bennettii',
         'authority': 'Waterhouse, 1837',
         'commonNameEnglish': "Bennett's Chinchilla Rat",
         'taxonomyType': 'biological',
         'uniqueIdentifier': '7a6c93a5-bdf7-4182-82f9-7a67d23f7fe1'}
        """
        df = pd.DataFrame(d['data'])
        df.to_csv(wi_taxonomy_csv_path,index=False)

    return df


def read_lila_taxonomy_mapping(metadata_dir, force_download=False):
    """
    Reads the LILA taxonomy mapping file, downloading the .csv file if necessary.

    Args:
        metadata_dir (str): folder to use for temporary LILA metadata files
        force_download (bool, optional): download the taxonomy mapping file
            even if the local file exists.

    Returns:
        pd.DataFrame: a DataFrame with one row per identification
    """

    p = urlparse(lila_taxonomy_mapping_url)
    taxonomy_filename = os.path.join(metadata_dir,os.path.basename(p.path))
    download_url(lila_taxonomy_mapping_url, taxonomy_filename,
                 force_download=force_download)

    df = pd.read_csv(lila_taxonomy_mapping_url)

    return df


def read_lila_metadata(metadata_dir, force_download=False):
    """
    Reads LILA metadata (URLs to each dataset), downloading the .csv file if necessary.

    Args:
        metadata_dir (str): folder to use for temporary LILA metadata files
        force_download (bool, optional): download the metadata file even if
            the local file exists.

    Returns:
        dict: a dict mapping dataset names (e.g. "Caltech Camera Traps") to dicts
        with keys corresponding to the headers in the .csv file, currently:

        - name
        - short_name
        - continent
        - country
        - region
        - image_base_url_relative
        - bbox_url_relative
        - image_base_url_gcp
        - metadata_url_gcp
        - bbox_url_gcp
        - image_base_url_aws
        - metadata_url_aws
        - bbox_url_aws
        - image_base_url_azure
        - metadata_url_azure
        - box_url_azure
        - mdv4_results_raw
        - mdv5b_results_raw
        - md_results_with_rde
        - json_filename
    """

    # Put the master metadata file in the same folder where we're putting images
    p = urlparse(lila_metadata_url)
    metadata_filename = os.path.join(metadata_dir,os.path.basename(p.path))
    download_url(lila_metadata_url, metadata_filename, force_download=force_download)

    df = pd.read_csv(metadata_filename)

    records = df.to_dict('records')

    # Parse into a table keyed by dataset name
    metadata_table = {}

    # r = records[0]
    for r in records:
        if is_empty(r['name']):
            continue

        # Convert NaN's to None
        for k in r.keys():
            if is_empty(r[k]):
                r[k] = None

        metadata_table[r['name']] = r

    return metadata_table


def read_lila_all_images_file(metadata_dir, force_download=False):
    """
    Downloads if necessary - then unzips if necessary - the .csv file with label mappings for
    all LILA files, and opens the resulting .csv file as a Pandas DataFrame.

    Args:
        metadata_dir (str): folder to use for temporary LILA metadata files
        force_download (bool, optional): download the metadata file even if
            the local file exists.

    Returns:
        pd.DataFrame: a DataFrame containing one row per identification in a LILA camera trap image
    """

    p = urlparse(lila_all_images_url)
    lila_all_images_zip_filename = os.path.join(metadata_dir,os.path.basename(p.path))
    download_url(lila_all_images_url, lila_all_images_zip_filename,
                 force_download=force_download)

    with zipfile.ZipFile(lila_all_images_zip_filename,'r') as z:
        files = z.namelist()
    assert len(files) == 1

    unzipped_csv_filename = os.path.join(metadata_dir,files[0])
    if not os.path.isfile(unzipped_csv_filename):
        unzip_file(lila_all_images_zip_filename,metadata_dir)
    else:
        print('{} already unzipped'.format(unzipped_csv_filename))

    df = pd.read_csv(unzipped_csv_filename)

    return df


def read_metadata_file_for_dataset(ds_name,
                                   metadata_dir,
                                   metadata_table=None,
                                   json_url=None,
                                   preferred_cloud='gcp',
                                   force_download=False):
    """
    Downloads if necessary - then unzips if necessary - the .json file for a specific dataset.

    Args:
        ds_name (str): the name of the dataset for which you want to retrieve metadata (e.g.
            "Caltech Camera Traps")
        metadata_dir (str): folder to use for temporary LILA metadata files
        metadata_table (dict, optional): an optional dictionary already loaded via
            read_lila_metadata()
        json_url (str, optional): the URL of the metadata file, if None will be retrieved
            via read_lila_metadata()
        preferred_cloud (str, optional): 'gcp' (default), 'azure', or 'aws'
        force_download (bool, optional): download the metadata file even if
            the local file exists.

    Returns:
        str: the .json filename on the local disk

    """

    if preferred_cloud is None:
        preferred_cloud = 'gcp'

    assert preferred_cloud in lila_base_urls.keys()

    if json_url is None:

        if metadata_table is None:
            metadata_table = read_lila_metadata(metadata_dir)

        json_url = metadata_table[ds_name]['metadata_url_' + preferred_cloud]

    p = urlparse(json_url)
    json_filename = os.path.join(metadata_dir,os.path.basename(p.path))
    download_url(json_url, json_filename, force_download=force_download)

    # Unzip if necessary
    if json_filename.endswith('.zip'):

        with zipfile.ZipFile(json_filename,'r') as z:
            files = z.namelist()
        assert len(files) == 1
        unzipped_json_filename = os.path.join(metadata_dir,files[0])
        if not os.path.isfile(unzipped_json_filename):
            unzip_file(json_filename,metadata_dir)
        else:
            print('{} already unzipped'.format(unzipped_json_filename))
        json_filename = unzipped_json_filename

    return json_filename


#%% Interactive test driver

if False:

    pass

    #%% Verify that all base URLs exist

    # LILA camera trap primary metadata file
    urls = (lila_metadata_url,
            lila_taxonomy_mapping_url,
            lila_all_images_url,
            wildlife_insights_taxonomy_url)

    from megadetector.utils import url_utils

    status_codes = url_utils.test_urls(urls,timeout=2.0)
    assert all([code == 200 for code in status_codes])


    #%% Verify that the metadata URLs exist for individual datasets

    metadata_dir = os.path.expanduser('~/lila/metadata')

    dataset_metadata = read_lila_metadata(metadata_dir)

    urls_to_test = []

    # ds_name = next(iter(dataset_metadata.keys()))
    for ds_name in dataset_metadata.keys():

        ds_info = dataset_metadata[ds_name]
        for cloud_name in lila_base_urls.keys():
            urls_to_test.append(ds_info['metadata_url_' + cloud_name])
            if ds_info['bbox_url_relative'] is not None:
                urls_to_test.append(ds_info['bbox_url_' + cloud_name])

    status_codes = url_utils.test_urls(urls_to_test,
                                       error_on_failure=True,
                                       n_workers=10,
                                       pool_type='process',
                                       timeout=2.0)
    assert all([code == 200 for code in status_codes])

